#!/usr/bin/python
# 
# Generate rise and set times for the Sun and Moon using PyEphem
#

# Import functions

from ephem import *
from datetime import datetime, timedelta
from pytz import timezone, all_timezones
from math import floor, pi
from re import *
from time import gmtime
import cgi

# Use these lines to debug the script

import cgitb
cgitb.enable()

# Default parameters for Moore Observatory
moorelat     = '38:20:40'; 
moorelong    = '-85:31:44'; 
adate        = (2009,10,15,12,0,0)
mooremajor   = 'US'
moorezone    = 'Eastern'

latitude     = moorelat
longitude    = moorelong
dfltmajor    = mooremajor
dfltzone     = moorezone


# Default parameters for Mt. Kent Observatory

mtkentlat    = '-27:47:52.3';
mtkentlong   = '151:51:19.5'; 
mtkentmajor  = 'Australia'
mtkentzone   = 'Brisbane'


dfltmajor2   =  mtkentmajor
dfltzone2    =  mtkentzone

action       = 'majors'

HUGEDAYS     = 1000*365.25

lunarperiod  = 29.530588853

########################################################################
# Converts an ephem UTC date to a local time hh:mm
def ut2ltstr(adate):
  dtuple = list(adate.tuple())
  dtuple[5] = int(floor(dtuple[5]))
  ut     = apply(datetime,dtuple,{'tzinfo':timezone('UTC')})
  lt     = ut.astimezone(tz)
  return lt.strftime('%H:%M')


########################################################################
# Converts an ephem UTC date to a local time yyyy mmm dd hh:mm:ss
def ut2ltstr2(adate):
  dtuple = list(date(adate).tuple())
  dtuple[5] = int(floor(dtuple[5]))
  ut     = apply(datetime,dtuple,{'tzinfo':timezone('UTC')})
  lt     = ut.astimezone(tz)
  return lt.strftime('%Y %b %d %H:%M:%S')

########################################################################
# Converts an ephem UTC date to a local date string yyyy mmm dd
def ut3ltstr(adate):
  dtuple = list(adate.tuple())
  dtuple[5] = int(floor(dtuple[5]))
  ut     = apply(datetime,dtuple,{'tzinfo':timezone('UTC')})
  lt     = ut.astimezone(tz)
  return lt.strftime('%Y %b %d')

########################################################################
def GetTimeZones():
  global majors, minors
  re1  = compile(r'(\w+)/([\w+-_]+)')
  re2  = compile(r'([\w+-]+)')
  tzs  = all_timezones
  lth  = len(tzs); i = 0;
  lastMaj = []
  while i<lth:
    tz = tzs[i]
    amatch = re1.search(tz)
    if amatch:
      maj = amatch.group(1)
      if maj != lastMaj:
        lastMaj = maj
        majors.append(maj)
      if maj == reqmajor:
        minors.append(amatch.group(2))
    else:
      amatch = re2.search(tz)
      if (amatch) and (reqmajor == 'Misc'):
        minors.append(amatch.group(1))
    i += 1
  majors.append('Misc')
 

########################################################################
# Set adate to timezone time for today at this time
def setadate():
  global tzstr
  tzstr = dfltmajor + '/' + dfltzone
  utl    = list(gmtime())
  utl    = utl[0:6]
  utzone = timezone('UTC')
  ut     = apply(datetime,utl,{'tzinfo': utzone})
  ltzone = timezone(tzstr)
  lt     = ut.astimezone(ltzone)
  ltstr  = lt.strftime('%Y %m %d %H %M %S')     # %H %M %S added Dec 2007
  lst    = ltstr.split()
  for i in range(6):
    lst[i] = int(lst[i])
  return lst

########################################################################


# Moore Observatory

reqmajor     = dfltmajor
reqzone      = dfltzone

# Retrieve the timezones
majors   = []; minors = []; 
GetTimeZones()

# Generate the location

locn     = Observer()
locn.lat = moorelat; 
locn.long = moorelong;
dfltmajor = mooremajor
dfltzone  = moorezone

# Generate full timezone string

adate = setadate() 


tz       = timezone(tzstr)
tt       = apply(datetime,adate)              
lt       = tz.localize(tt,is_dst = True)      
ldatestr = lt.strftime('%Y %b %d  %H:%M:%S')   
ndatestr = lt.strftime('%Y %m %d %H %M %S')   
offsstr  = lt.strftime('%z')
prev     = lt - timedelta(hours = 24)
prevstr  = prev.strftime('%Y %b %d')
ut       = lt.astimezone(timezone('UTC'))
utstr    = ut.strftime('%Y/%m/%d %H:%M:%S')
utstr2   = ut.strftime('%Y %b %d %H:%M:%S')

# Generate the data

sset = []; 
srise = []; 
dawn = []; 
dusk = [];
strans = []; 
azsrise = []; 
altstrans = []; 
azsset = [];
newmoon = [];
firstquartermoon = [];
fullmoon =[];
lastquartermoon = [];


datestr = ut.strftime('%Y/%m/%d')
locn.date    = datestr
sun = Sun() 
moon = Moon()

# Do rise and set ut

locn.horizon =     '0.0'                             

srise = locn.next_rising(sun)
azsrise = sun.az

sset = locn.next_setting(sun)
azsset = sun.az 

strans = locn.next_transit(sun)
altstrans = sun.alt

mrise = locn.next_rising(moon)
azmrise = moon.az

mtrans = locn.next_transit(moon)
altmtrans = moon.alt

mset = locn.next_setting(moon)
azmset = moon.az 

# Do astro twilight ut

locn.horizon = '-18.0'

dawn = locn.next_rising(sun)
dusk = locn.next_setting(sun)

# Get moon phase information

newmoon    = next_new_moon(utstr)
firstmoon  = next_first_quarter_moon(utstr)
fullmoon   = next_full_moon(utstr)
lastmoon   = next_last_quarter_moon(utstr)

# Output local times for this zone

ut0 = date(utstr)
dawn = ut2ltstr2(dawn)
dusk = ut2ltstr2(dusk)
srise = ut2ltstr2(srise)
strans = ut2ltstr2(strans)
sset = ut2ltstr2(sset)
newmoon = ut3ltstr(newmoon) 
firstmoon = ut3ltstr(firstmoon)
fullmoon = ut3ltstr(fullmoon)
lastmoon = ut3ltstr(lastmoon)



#print 'Content-type: text/html'
#print

#print 'Sun Set:              ' + str(sset) 
#print 'Sun Rise:             ' + str(srise) 
#print 'Dawn:                 ' + str(dawn) 
#print 'Dusk:                 ' + str(dusk) 
#print 'Sun Transit:          ' + str(strans) 
#print 'Sun Rise Azimuth:     ' + str(azsrise)
#print 'Sun Transit Altitude: ' + str(altstrans)
#print 'Sun Set Azimuth:      ' + str(azsset) 

#print 'New Moon:      ' + str(newmoon)
#print 'First Quarter: ' + str(firstmoon)
#print 'Full:          ' + str(fullmoon)
#print 'Last Quarter:  ' + str(lastmoon)

filename = "/home/www/htdocs/obstools/moore_sunrise.html"
html_lines = []
html_file = open(filename,"w")
html_lines.append('')
html_lines.append('<br>\n')
html_lines.append('Universal Time:       ' + str(utstr2) + '\n')
html_lines.append('<br>\n')
html_lines.append('Local zone:           ' + str(tzstr) + '\n')
html_lines.append('<br>\n')
html_lines.append('Local time:           ' + str(ldatestr) + '\n')
html_lines.append('<br>\n')
html_lines.append('<br>\n')
html_lines.append('Dawn:                 ' + str(dawn) + '\n') 
html_lines.append('<br>\n')
html_lines.append('Sun Rise:             ' + str(srise) + '\n') 
html_lines.append('<br>\n')
html_lines.append('Sun Transit:          ' + str(strans) + '\n') 
html_lines.append('<br>\n')
html_lines.append('Sun Set:              ' + str(sset) + '\n')
html_lines.append('<br>\n')
html_lines.append('Dusk:                 ' + str(dusk) + '\n') 
html_lines.append('<br>\n')
html_lines.append('<br>\n')
html_lines.append('Sun Rise Azimuth:     ' + str(azsrise) + '\n')
html_lines.append('<br>\n')
html_lines.append('Sun Transit Altitude: ' + str(altstrans) + '\n')
html_lines.append('<br>\n')
html_lines.append('Sun Set Azimuth:      ' + str(azsset) + '\n') 
html_lines.append('<br>\n')
html_lines.append('<br>\n')
html_lines.append('New Moon:      ' + str(newmoon) + '\n')
html_lines.append('<br>\n')
html_lines.append('First Quarter: ' + str(firstmoon) + '\n')
html_lines.append('<br>\n')
html_lines.append('Full Moon:          ' + str(fullmoon) + '\n')
html_lines.append('<br>\n')
html_lines.append('Last Quarter:  ' + str(lastmoon) + '\n')
html_lines.append('<br>\n')
html_file.writelines(html_lines)
html_file.close()



# Mt. Kent Observatory

reqmajor     = mtkentmajor
reqzone      = mtkentzone

# Retrieve the timezones
majors   = []; minors = []; 
GetTimeZones()

# Generate the location

locn     = Observer()
locn.lat = mtkentlat; 
locn.long = mtkentlong;
dfltmajor = mtkentmajor
dfltzone  = mtkentzone

# Generate full timezone string

adate = setadate() 

tz       = timezone(tzstr)
tt       = apply(datetime,adate)              
lt       = tz.localize(tt,is_dst = True)      
ldatestr = lt.strftime('%Y %b %d %H %M %S')   
ndatestr = lt.strftime('%Y %m %d %H %M %S')   
offsstr  = lt.strftime('%z')
prev     = lt - timedelta(hours = 24)
prevstr  = prev.strftime('%Y %b %d')
ut       = lt.astimezone(timezone('UTC'))
utstr    = ut.strftime('%Y/%m/%d %H:%M:%S')

# Generate the data

sset = []; 
srise = []; 
dawn = []; 
dusk = [];
strans = []; 
azsrise = []; 
altstrans = []; 
azsset = [];
newmoon = [];
firstquartermoon = [];
fullmoon =[];
lastquartermoon = [];


datestr = ut.strftime('%Y/%m/%d')
locn.date    = datestr
sun = Sun() 
moon = Moon()

# Do rise and set ut

locn.horizon =     '0.0'                             

srise = locn.next_rising(sun)
azsrise = sun.az

sset = locn.next_setting(sun)
azsset = sun.az 

strans = locn.next_transit(sun)
altstrans = sun.alt

mrise = locn.next_rising(moon)
azmrise = moon.az

mtrans = locn.next_transit(moon)
altmtrans = moon.alt

mset = locn.next_setting(moon)
azmset = moon.az 

# Do astro twilight ut

locn.horizon = '-18.0'

dawn = locn.next_rising(sun)
dusk = locn.next_setting(sun)

# Get moon phase information ut

newmoon    = next_new_moon(utstr)
firstmoon  = next_first_quarter_moon(utstr)
fullmoon   = next_full_moon(utstr)
lastmoon   = next_last_quarter_moon(utstr)

# Output local times for this zone

ut0 = date(utstr)
dawn = ut2ltstr2(dawn)
dusk = ut2ltstr2(dusk)
srise = ut2ltstr2(srise)
strans = ut2ltstr2(strans)
sset = ut2ltstr2(sset)
newmoon = ut3ltstr(newmoon) 
firstmoon = ut3ltstr(firstmoon)
fullmoon = ut3ltstr(fullmoon)
lastmoon = ut3ltstr(lastmoon)

#print 'Content-type: text/html'
#print

#print 'Sun Set:              ' + str(sset) 
#print 'Sun Rise:             ' + str(srise) 
#print 'Dawn:                 ' + str(dawn) 
#print 'Dusk:                 ' + str(dusk) 
#print 'Sun Transit:          ' + str(strans) 
#print 'Sun Rise Azimuth:     ' + str(azsrise)
#print 'Sun Transit Altitude: ' + str(altstrans)
#print 'Sun Set Azimuth:      ' + str(azsset) 

#print 'New Moon:      ' + str(newmoon)
#print 'First Quarter: ' + str(firstmoon)
#print 'Full:          ' + str(fullmoon)
#print 'Last Quarter:  ' + str(lastmoon)

filename = "/home/www/htdocs/obstools/mtkent_sunrise.html"
html_lines = []
html_file = open(filename,"w")
html_lines.append('')
html_lines.append('<br>\n')
html_lines.append('Universal Time:       ' + str(utstr2) + '\n')
html_lines.append('<br>\n')
html_lines.append('Local zone:           ' + str(tzstr) + '\n')
html_lines.append('<br>\n')
html_lines.append('Local time:           ' + str(lt) + '\n')
html_lines.append('<br>\n')
html_lines.append('<br>\n')
html_lines.append('Dawn:                 ' + str(dawn) + '\n') 
html_lines.append('<br>\n')
html_lines.append('Sun Rise:             ' + str(srise) + '\n') 
html_lines.append('<br>\n')
html_lines.append('Sun Transit:          ' + str(strans) + '\n') 
html_lines.append('<br>\n')
html_lines.append('Sun Set:              ' + str(sset) + '\n')
html_lines.append('<br>\n')
html_lines.append('Dusk:                 ' + str(dusk) + '\n') 
html_lines.append('<br>\n')
html_lines.append('<br>\n')
html_lines.append('Sun Rise Azimuth:     ' + str(azsrise) + '\n')
html_lines.append('<br>\n')
html_lines.append('Sun Transit Altitude: ' + str(altstrans) + '\n')
html_lines.append('<br>\n')
html_lines.append('Sun Set Azimuth:      ' + str(azsset) + '\n') 
html_lines.append('<br>\n')
html_lines.append('<br>\n')
html_lines.append('New Moon:      ' + str(newmoon) + '\n')
html_lines.append('<br>\n')
html_lines.append('First Quarter: ' + str(firstmoon) + '\n')
html_lines.append('<br>\n')
html_lines.append('Full Moon:          ' + str(fullmoon) + '\n')
html_lines.append('<br>\n')
html_lines.append('Last Quarter:  ' + str(lastmoon) + '\n')
html_lines.append('<br>\n')
html_file.writelines(html_lines)
html_file.close()




print "Content-type: text/html"

print """

<br><br><br><br><br><br>
<center>
<br>
Welcome to Shared Skies
<br>
<br>
Select an Observatory
</center>
"""
