/* -------------------------------------------------------------------------- */
/* -                   Astronomical Telescope Control                       - */
/* -                        Motif User Interface                            - */
/* -                        INDI Client for XmTel                           - */
/* -------------------------------------------------------------------------- */
/*                                                                            */
/* Copyright (c) 2008-2012 John Kielkopf                                      */
/* kielkopf@louisville.edu                                                    */
/*                                                                            */
/* This file is part of XmTel.                                                */
/*                                                                            */
/* Distributed under the terms of the General Public License (see LICENSE)    */ 
/*                                                                            */
/* Date: July 22, 2012                                                        */
/* Version: 6.0                                                               */
/*                                                                            */
/* About epochs ...                                                           */
/*   XmTel uses epoch of date (EOD) coordinates for telescope operation       */
/*   Telescope coordinates are provided read-only by the INDI server          */
/*     both EOD and epoch 2000.0 but EOD are the ones used to point           */
/*   User inteface display may be toggled between epochs                      */
/*   Coordinates for a target may be entered epoch 2000.0 or EOD              */
/*   Epoch conversion from queue includes proper motion to EOD if known       */
/*   Target is always sent as a EOD request                                   */
/*   Target display in epoch 2000 is the report from the server               */
/*                                                                            */
/* See TELCOMMANDS for the xml parameters used by xmtel                       */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <ctype.h>
#include <netinet/in.h>
#include <netdb.h>


#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/ToggleB.h>
#include <Xm/FileSB.h>
#include <Xm/PushB.h>
#include <Xm/Label.h>
#include <Xm/RowColumn.h> 
#include <Xm/CascadeB.h>
#include <Xm/MessageB.h>
#include <Xm/Separator.h>
#include <Xm/SelectioB.h>
#include <Xm/Text.h>
#include <Xm/List.h>

#include "lilxml.h"
#include "indiapi.h"
#include "xmtel.h"

XtAppContext context;
XmStringCharSet char_set=XmSTRING_DEFAULT_CHARSET;

/* Widgets for the menu bar */

Widget menu_bar;
Widget file_menu;
Widget newqueue_item;
Widget newlog_item;
Widget exit_item;
Widget edit_menu;
Widget editqueue_item;
Widget editlog_item;
Widget pointing_menu;
Widget p_options_toggle_1;
Widget p_options_toggle_2;
Widget p_options_toggle_4;
Widget p_options_toggle_8;
Widget p_options_toggle_16;
Widget p_options_toggle_32;
Widget ref_menu;
Widget ref_target_item;
Widget ref_wcs_item;
Widget ref_clear_item;
Widget ref_save_item;
Widget ref_recall_item;
Widget ref_default_item;
Widget model_menu;
Widget model_edit_item;
Widget model_clear_item;
Widget model_save_item;
Widget model_recall_item;
Widget model_default_item;

Widget telescope_menu;
Widget focus_item;
Widget rotate_item;
Widget fan_item;
Widget heater_item;


/* Widgets for the panel */

Widget toplevel;
Widget dialog;
Widget button[3][3];
Widget command[12];
Widget toggle[4];
Widget form;
Widget sep_btns;
Widget sep_coords;
Widget sep_queue;
Widget sep_status;
Widget label_mark;
Widget label_ra;
Widget label_dec;
Widget label_epoch;
Widget mark_tel;
Widget mark_target;

Widget get_telra;
Widget get_teldec;
Widget get_telepoch;
Widget get_targetra;
Widget get_targetdec;
Widget get_targetepoch;

Widget select_logfile;
Widget read_queuefile;

Widget queue_area;
Widget message_area;
Widget radio_box;

/* User interface polling */

XtIntervalId poll_interval_id;
XtPointer poll_interval_data_ptr;
XtInputId indi_id;

Widget make_menu_item();     /* add a pushbutton item to the menu */
Widget make_menu_toggle();   /* add a toggle item to the menu     */
Widget make_menu();          /* create a menu on the menu bar     */

/* Widgets for the model_edit panel */

Widget model_edit;

/* Widgets for the focus panel */

Widget focus_control;     
Widget focus_in_button;
Widget focus_out_button;
Widget focus_ok_button;
Widget focus_count_label;
Widget focus_spd_radio_box;
Widget focus_spd_toggle[4];

/* Widgets for the rotate panel */

Widget rotate_control;     
Widget rotate_ok_button;
Widget rotate_count_label;
Widget rotate_spd_radio_box;
Widget rotate_spd_toggle[4];
Widget rotate_cmd_radio_box;
Widget rotate_cmd_toggle[4];

/* Widgets for the fan panel */

Widget fan_control;     
Widget fan_ok_button;
Widget fan_spd_radio_box;
Widget fan_spd_toggle[3];

/* Widgets for the heater panel */

Widget heater_control;     
Widget heater_ok_button;
Widget heater_pwr_radio_box;
Widget heater_pwr_toggle[3];


/* Labels for the guide buttons top to bottom then left to right */

static char btn_name[][6] = {
    "NoOp", 
    "E",
    "NoOp",
    "N",
    "NoOp",
    "S",
    "NoOp",
    "W",
    "NoOp"
};

/* Simple labels for the command area top to bottom, left to right */

static char cmd_name[][15] = {
    "Slew",
    "Save",
    "Track",
    "Recall",
    "Guide",  
    "Meridian",  
    "Stop",
    "Park"
};

/* Labels for the guide speed settings */

static char spd_name[][8] = {
    "Slew",
    "Find",
    "Center",
    "Guide"   
};

/* Labels for the focus speed settings */

static char focus_spd_name[][12] = {
     "8X",
     "4X",
     "2X",
     "1X" 
};   
 

/* Labels for the rotate speed settings */

static char rotate_spd_name[][12] = {
     "Fast",
     "Slow",
     "Sidereal"
};

/* Labels for the rotate command settings */

static char rotate_cmd_name[][12] = {
     "CW",
     "CCW",
     "Off"
};

/* Labels for the fan speed settings */

static char fan_spd_name[][12] = {
     "High",
     "Low",
     "Off"
};

/* Labels for the heater power settings */

static char heater_pwr_name[][12] = {
     "High",
     "Low",
     "Off"
};    

/* Message area string */

char message[300];


/* Menu functions */

void create_menus(Widget menu_bar);  /* Creates all menus for program         */
void menuCB();             /* Callback routine used for all menus             */
void setup_model_edit();   /* Create the unmanaged model edit panel           */
void model_editCB();       /* Callback function for model edit panel          */
void setup_focus();        /* Create the unmanaged focus panel                */
void focusCB();            /* Callback function for the focus panel           */
void focus_speedCB();      /* Callback function for focus speed toggles       */
void setup_rotate();       /* Create the unmanaged rotate panel               */
void rotateCB();           /* Callback function for the rotate panel          */
void rotate_speedCB();     /* Callback function for rotate speed toggles      */
void rotate_commandCB();   /* Callback function for rotate command toggles    */
void setup_fan();          /* Create the unmanaged fan panel                  */
void fanCB();              /* Callback function for the fan panel             */
void fan_speedCB();        /* Callback function for fan speed toggles         */
void setup_heater();       /* Create the unmanaged heater panel               */
void heaterCB();           /* Callback function for the heater panel          */
void heater_powerCB();     /* Callback function for heater power toggles      */

void dialogCB();           /* Callback function for the dialog box            */
void buttonCB();           /* Callback function for the paddle buttons        */
void speedCB();            /* Callback function for the slew speed buttons    */
void teldataCB();          /* Callback function for the data buttons          */
void commandCB();          /* Callback function for the command buttons       */
void selectlogCB();        /* Callback function for selecting the log file    */
void readqueueCB();        /* Callback function for opening the queue file    */
void selectqueueCB();      /* Callback function for selecting a queue entry   */

/* Telescope functions */

void show_telescope_coordinates();      /* Display latest telescope coordinates */
void show_target_coordinates();         /* Display latest target coordinates  */
void show_model_parameters();           /* Display latest model parameters */
void show_focus();                      /* Display the focus reading */
void show_rotation();                   /* Display the rotation reading */
void show_temperature();                /* Display the temperature */
void show_guide_status();               /* Display the guide status */

/* Message area display */

void show_message();                  /* Update message area with string message */

/* Local log and queue operations */

void save_coordinates();                /* Save a log entry and add to the history */
void recall_coordinates();              /* Read the next previous history entry */
void read_queue();                      /* Read queue file into memory */

/* User interface RA and Dec direct entry */

void new_target_ra();                   /* User input of target RA */
void new_target_dec();                  /* User input of target Dec */

/* Interface to a local sky display */

void link_fifos();                    /* Startup fifo link routine */
void unlink_fifos();                  /* Shutdown fifo link routine */
void mark_xephem_telescope();         /* Export telescope coordinates to xephem */
void read_xephem_target();            /* Read target coordinates from xephem goto fifo */
void mark_xephem_target();            /* Wxport target coordinates to xephem */

/* Convert string deg:min:sec to double */

int dmstod (char *instr, double *datap);

/* Convert dms double to string deg:min:sec */

void dtodms (char *outstr, double *dmsp); 

/* Site Parameters (Actual values are defined in the driver ) */

double SiteLatitude = 85.5300;       /* Latitude in degrees + north */  
double SiteLongitude = 38.3334;      /* Longitude in degrees + west */  
double SiteAltitude = 230.0;         /* Altitude in meters */
double SitePressure = 760.0;         /* Atmospheric pressure in Torr */
double SiteTemperature = 20.0;       /* Local atmospheric temperature  C */

/* Telescope control */

char   telescope_device[64];
char   telescope_parameters[4096];
char   telescope_set_prm[1024];

/* Telescope parameters provided read-only to this client */

double telfocus=0.0;          /* Focuser position         */
double telrotation=0.0;       /* Rotator angle            */ 
double teltemperature=0.0;    /* OTA temperature sensor   */
double telra;                 /* Epoch of date!!          */
double teldec;                /* Epoch of date!!          */
double telra2;                /* Epoch 2000.0             */
double teldec2;               /* Epoch 2000.0             */  

/* Target parameters provided read-write to this client   */ 

double targetra;              /* Epoch of date!!          */
double targetdec;             /* Epoch of date!!          */
double targetra2;             /* Epoch 2000.0             */
double targetdec2;            /* Epoch 2000.0             */  

/* Model parameters provided read-write to this client */

double modelha1;              /* Pixels/hour HA drift */
double modeldec1;             /* Pixels/hour dec drift */

/* User interface flags and variables  */

int display_telepoch=EOD;             /* Display epoch for telescope */
int display_targetepoch=EOD;          /* Display epoch for target  */
int pmodel=RAW;                       /* Pointing model default to raw data */
int telspd=FIND;                      /* Drive speed */
int telflag=FALSE;                    /* Telescope connection flag */
int target=RA;                        /* Flag for target field input */
int trackcmd;                         /* Flag to indicate tracking state */
int refcmd;                           /* Flag to indicate reference state */
int targetcmd;                        /* Flag to indicate target set state */
int slewcmd;                          /* Flag to indicate current slew activity */
int guidecmd;                         /* Flag to indicate guiding state */
int teldir;                           /* Flag to indicate current slew direction */
int gotocmd=FALSE;                    /* Flag to indicate status of a goto request */
int focuscmd=FOCUSCMDOFF;             /* Flag to indicate current focus activity */
int focusspd=FOCUSSPD4;               /* Flag to indicate current focus speed */
int rotatecmd=ROTATECMDOFF;           /* Flag to indicate rotator state */
int rotatespd=ROTATESPDSLOW;          /* Flag to indicate rotator speed */
int fancmd=FANCMDOFF;                 /* Flag to indicate fan state */
int heatercmd=HEATERCMDOFF;           /* Flag to indicate heater state */

/* Files */

int fd_fifo_in, fd_fifo_out;           /* FIFO file descriptors */
FILE *fp_log;                          /* Log file pointer      */
static char *logfile;                  /* Log name */
FILE *fp_queue;                        /* Queue file pointer */
static char *queuefile;                /* Queue name */

/* External commands */

static char *editor;                   /* External queue and log editor */

/* Buffers */

char buf[256];                         /* General purpose character buffer */

/* Queue */

int nqueue = 0;                        /* Number of entries in the queue */
int queuechoice = 0;                   /* Serial number of selected entry */

/* History */

/* The history catalog is a ring buffer starting at 0 for the first entry. */
/* There are nhistory entries in use (<=100). */

int nhistory = 0;                      /* Number of history entries (max 100) */
int last_history_saved = -1;           /* Location for last history saved */
int last_history_recalled = -1;        /* Location for last history recalled */

typedef struct                         /* Catalog structure assuming EOD */
{
  char name[80];
  char desc[80];
  double ra;
  double dec;
  double mag;
} catalog;

/* Create the space for catalogs.  Must keep track not to overrun. */
/* Allow one extra entry and watch for maximum number in operations. */
/* History is indexed from 1 for first entry, so 0th entry is not used. */

catalog queue[10001];                  /* This holds the observing queue */
catalog history[101];                  /* This holds the saved coordinates */

/* Flags */

char option_state;

/* Indi functions */

void indiCB (XtPointer client, int *fdp, XtInputId *idp);
static int launch_indi_command (XMLEle *root, char errmsg[]);
static int set_indi_number (XMLEle *root, char errmsg[]);
static int set_indi_text (XMLEle *root, char errmsg[]);
static int set_indi_switch (XMLEle *root, char errmsg[]);
static int set_indi_light (XMLEle *root, char errmsg[]);
static int set_indi_blob (XMLEle *root, char errmsg[]);
static int set_indi_message (XMLEle *root, char errmsg[]);

int send_indi_number (FILE *fp, char *dev, char *prm, double value);
int send_indi_text (FILE *fp, char *dev, char *prm, char *txt);
int send_indi_switch (FILE *fp, char *dev, char *prm, int state);

FILE *indi_fp;
int indi_socket;

LilXML *lillp;

char host_def[] = "localhost";   /* Default host name */
char *host = host_def;           /* Working host name */

int port = INDIPORT;		 /* Working port number */
int diagnostics = DIAGNOSTICS;   /* Diagnostic level set false to turn off */
int connect_indi(void);
int disconnect_indi(void);

/* Corrections for proper motion, precession, aberration, and nutation */

extern void Apparent(double *ra, double *dec, int dirflag);  
extern void PrecessToEOD(double epoch, double  *ra, double  *dec);
extern void PrecessToEpoch(double epoch, double  *ra, double  *dec);
extern void ProperMotion(double epoch, double *ra, double *dec, 
  double pm_ra, double pm_dec); 


/* End of global definitions */

int main(argc,argv)
  int argc; 
  char *argv[];
{
  Arg al[10];
  int ac;
  unsigned long x,y,btn_number,cmd_number;
  EventMask mask;
  
  /* No command line arguments needed */
  
  sprintf(telescope_device,"tel");  

  /* Make contact with the server */
  
  telflag = FALSE;
  
  telflag = connect_indi();
  if (telflag == FALSE)
  {
    fprintf(stderr,"INDI server is not available\n");
    exit(1);
  }
  
  fprintf(stderr,"INDI server connected\n");
  
  /* Create parser context */
  
  if (lillp)
  {
    delLilXML (lillp);
  }
  lillp = newLilXML();  

  /* Define the defaults for log and queue files */
  
  logfile = (char *) malloc (MAXPATHLEN);
  strcpy(logfile,LOGFILE);
  queuefile = (char *) malloc (MAXPATHLEN);
  strcpy(queuefile,QUEUEFILE);
  editor = (char *) malloc (MAXPATHLEN);
  strcpy(editor,XMTEL_EDITOR);

  /* Start the fifo communications */
  /* Start XEphem after XmTel is running */
  
  link_fifos();
 
  /* Create the toplevel shell */

  toplevel = XtAppInitialize(&context,"",NULL,0,&argc,argv,
      NULL,NULL,0);

  /* Set the default size of the window */ 

  ac=0;
  XtSetArg(al[ac],XmNwidth,400); ac++;
  XtSetArg(al[ac],XmNheight,700); ac++;
  XtSetValues(toplevel,al,ac);

  /* Create a form widget */

  ac=0;
  form=XmCreateForm(toplevel,"form",al,ac);
  XtManageChild(form);

  /* Create the menu bar and attach it to the form */
  
  ac=0;
  XtSetArg(al[ac],XmNtopAttachment,XmATTACH_FORM); ac++;
  XtSetArg(al[ac],XmNrightAttachment,XmATTACH_FORM); ac++;
  XtSetArg(al[ac],XmNleftAttachment,XmATTACH_FORM); ac++;
  menu_bar=XmCreateMenuBar(form,"menu_bar",al,ac);
  XtManageChild(menu_bar);

  /* Note that locations of buttons are in percent of window dimension */

  /* Create a radio box to hold the speed toggles and attach it to form */
  
  ac=0;
  XtSetArg(al[ac],XmNorientation,XmVERTICAL); ac++;
  XtSetArg(al[ac],XmNtopAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,12); ac++; 
  XtSetArg(al[ac],XmNleftAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,5); ac++; 
  XtSetArg(al[ac],XmNrightAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,24); ac++; 
  radio_box=XmCreateRadioBox(form,"radio_box",al,ac);
  XtManageChild(radio_box);
  
  /* Create speed toggles */
  
  for (x=0; x<4; x++)
  {
    ac=0;
    XtSetArg(al[ac],XmNlabelString,
      XmStringCreate(spd_name[x],char_set)); ac++;
    toggle[x]=XmCreateToggleButton(radio_box,"toggle",al,ac);
    XtManageChild(toggle[x]);
    XtAddCallback (toggle[x], XmNvalueChangedCallback, speedCB, (XtPointer) x);
  }     
 
  /* Create the slew speed togglebuttons with find speed selected */

  XmToggleButtonSetState (toggle[1], True, False);
  
  /* Set up the slew buttons, add event handlers, attach to form. */

  for (x=0; x<3; x++)
  {
    for (y=0; y<3; y++)
    {
      btn_number=3*x+y;
      if ( (btn_number==1) | (btn_number==3) | (btn_number==5) | (btn_number==7) ) 
      {
        ac=0;
        XtSetArg(al[ac],XmNlabelString,
            XmStringCreate(btn_name[btn_number],char_set)); ac++;
        XtSetArg(al[ac],XmNleftAttachment,
            XmATTACH_POSITION); ac++;
        XtSetArg(al[ac],XmNleftPosition,28+x*15); ac++;
        XtSetArg(al[ac],XmNrightAttachment,
            XmATTACH_POSITION); ac++;
        XtSetArg(al[ac],XmNrightPosition,42+x*15); ac++;
        XtSetArg(al[ac],XmNtopAttachment,
            XmATTACH_POSITION); ac++;
        XtSetArg(al[ac],XmNtopPosition,6+y*9); ac++;
        XtSetArg(al[ac],XmNbottomAttachment,
            XmATTACH_POSITION); ac++;
        XtSetArg(al[ac],XmNbottomPosition, 15+y*9); ac++;
        button[x][y]=XmCreatePushButton(form,"label",al,ac);
        mask = ButtonReleaseMask | ButtonPressMask;
        XtAddEventHandler(button[x][y],mask,False,
            buttonCB,(XtPointer) btn_number);              
        XtManageChild(button[x][y]);
      }
    }
  }

  /* Create a separator widget and attach it to the form */

  ac=0;
  XtSetArg(al[ac],XmNtopAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,35); ac++;
  XtSetArg(al[ac],XmNrightAttachment,XmATTACH_FORM); ac++; 
  XtSetArg(al[ac],XmNleftAttachment,XmATTACH_FORM); ac++;  
  XtSetArg(al[ac],XmNbottomAttachment,XmATTACH_NONE); ac++;  
  sep_btns=XmCreateSeparator(form,"sep",al,ac);
  XtManageChild(sep_btns);   
 
  /* Build the telescope and target coordinates display area */
  
  /* Create the mark label widget and attach it to the form */

  ac=0;
  XtSetArg(al[ac],XmNlabelString,
      XmStringCreate("Mark",char_set)); ac++;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,13); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,31); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,36); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,40); ac++;
  label_mark=XmCreateLabel(form,"label",al,ac);
  XtManageChild(label_mark); 
  
  /* Create the ra label widget and attach it to the form */

  ac=0;
  XtSetArg(al[ac],XmNlabelString,
      XmStringCreate("RA",char_set)); ac++;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,37); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,45); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,36); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,40); ac++;
  label_ra=XmCreateLabel(form,"label",al,ac);
  XtManageChild(label_ra); 

  /* Create the dec label widget and attach it to the form */

  ac=0;
  XtSetArg(al[ac],XmNlabelString,
      XmStringCreate("Dec",char_set)); ac++;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,55); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,63); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,36); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,40); ac++;
  label_dec=XmCreateLabel(form,"label",al,ac);
  XtManageChild(label_dec);    

  /* Create the epoch label widget and attach it to the form */

  ac=0;
  XtSetArg(al[ac],XmNlabelString,
      XmStringCreate("Epoch",char_set)); ac++;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,72); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,80); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,36); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,40); ac++;
  label_epoch=XmCreateLabel(form,"label",al,ac);
  XtManageChild(label_epoch);    

  /* Create the telescope label widget and attach it to the form */

  cmd_number = 0;
  ac=0;
  XtSetArg(al[ac],XmNlabelString,
      XmStringCreate("Telescope",char_set)); ac++;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,13); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,31); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,40); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,44); ac++;
  mark_tel=XmCreatePushButton(form,"label",al,ac);
  XtManageChild(mark_tel);
  XtAddCallback(mark_tel,XmNactivateCallback,
                teldataCB,(XtPointer) cmd_number);     
       
  /* Create the target label widget and attach it to the form */

  cmd_number = 1;
  ac=0;
  XtSetArg(al[ac],XmNlabelString,
      XmStringCreate("Target",char_set)); ac++;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,13); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,31); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,44); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,48); ac++;
  mark_target=XmCreatePushButton(form,"label",al,ac);
  XtManageChild(mark_target);    
  XtAddCallback(mark_target,XmNactivateCallback,
                teldataCB,(XtPointer) cmd_number);     
 
  /* Create a telescope ra widget and attach it to the form */

  cmd_number = 2;    
  ac=0;
  XtSetArg(al[ac],XmNlabelString,
      XmStringCreate("00:00:00",char_set)); ac++;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,33); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,49); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,40); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,44); ac++;
  get_telra=XmCreatePushButton(form,"label",al,ac);
  XtManageChild(get_telra);
  XtAddCallback(get_telra,XmNactivateCallback,
                teldataCB,(XtPointer) cmd_number);     
 
  /* Create a telescope dec widget and attach it to the form */

  cmd_number = 3;    
  ac=0;
  XtSetArg(al[ac],XmNlabelString,
      XmStringCreate("00:00:00",char_set)); ac++;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,51); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,67); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,40); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,44); ac++;
  get_teldec=XmCreatePushButton(form,"label",al,ac);
  XtManageChild(get_teldec);
  XtAddCallback(get_teldec,XmNactivateCallback,
                teldataCB,(XtPointer) cmd_number);    

  /* Create a telescope epoch widget and attach it to the form */

  cmd_number = 4;  
  ac=0;
  XtSetArg(al[ac],XmNlabelString,
      XmStringCreate("Now",char_set)); ac++;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,69); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,83); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,40); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,44); ac++;
  get_telepoch=XmCreatePushButton(form,"label",al,ac);
  XtManageChild(get_telepoch);
  XtAddCallback(get_telepoch,XmNactivateCallback,
                teldataCB,(XtPointer) cmd_number);     
  
  
  /* Create a target ra widget and attach it to the form */

  cmd_number = 5;  
  ac=0;
  XtSetArg(al[ac],XmNlabelString,
      XmStringCreate("00:00:00",char_set)); ac++;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,33); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,49); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,44); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,48); ac++;
  get_targetra=XmCreatePushButton(form,"label",al,ac);
  XtManageChild(get_targetra);
  XtAddCallback(get_targetra,XmNactivateCallback,
                teldataCB,(int *) cmd_number);
                        

  /* Create a target dec widget and attach it to the form */

  cmd_number = 6;  
  ac=0;
  XtSetArg(al[ac],XmNlabelString,
      XmStringCreate("00:00:00",char_set)); ac++;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,51); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,67); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,44); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,48); ac++;
  get_targetdec=XmCreatePushButton(form,"label",al,ac);
  XtManageChild(get_targetdec);
  XtAddCallback(get_targetdec,XmNactivateCallback,
                teldataCB,(XtPointer) cmd_number);    

  
  /* Create a target epoch widget and attach it to the form */

  cmd_number = 7;  
  ac=0;
  XtSetArg(al[ac],XmNlabelString,
      XmStringCreate("Now",char_set)); ac++;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,69); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,83); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,44); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,48); ac++;
  get_targetepoch=XmCreatePushButton(form,"label",al,ac);
  XtManageChild(get_targetepoch);
  XtAddCallback(get_targetepoch,XmNactivateCallback,
                teldataCB,(XtPointer) cmd_number);        
  

  /* Create a separator widget and attach it to the form */

  ac=0;
  XtSetArg(al[ac],XmNtopAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,50); ac++;
  XtSetArg(al[ac],XmNrightAttachment,XmATTACH_FORM); ac++; 
  XtSetArg(al[ac],XmNleftAttachment,XmATTACH_FORM); ac++;  
  XtSetArg(al[ac],XmNbottomAttachment,XmATTACH_NONE); ac++;      
  sep_coords=XmCreateSeparator(form,"sep",al,ac);
  XtManageChild(sep_coords);  
  

  /* Queue area */ 
  
  /* Create a scrolled queue list widget and attach it to the form */

  ac=0;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,8); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,92); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,52); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,63); ac++;
  queue_area=XmCreateScrolledList(form,"queue_area",al,ac);
  XtManageChild(queue_area);
  XtAddCallback(queue_area,XmNdefaultActionCallback,selectqueueCB,NULL);
  XtAddCallback(queue_area,XmNbrowseSelectionCallback,selectqueueCB,NULL);

    
  /* Create a separator widget and attach it to the form */

  ac=0;
  XtSetArg(al[ac],XmNtopAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,65); ac++;
  XtSetArg(al[ac],XmNrightAttachment,XmATTACH_FORM); ac++; 
  XtSetArg(al[ac],XmNleftAttachment,XmATTACH_FORM); ac++;  
  XtSetArg(al[ac],XmNbottomAttachment,XmATTACH_NONE); ac++;      
  sep_queue=XmCreateSeparator(form,"sep",al,ac);
  XtManageChild(sep_queue);   

  /* Set up the 8 command buttons and attach them to the form */

  for (x=0; x<4; x++)   
  for (y=0; y<2; y++)
  {
    cmd_number=2*x+y;
    ac=0;
    XtSetArg(al[ac],XmNlabelString,
      XmStringCreate(cmd_name[cmd_number],char_set)); ac++;
    XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
    XtSetArg(al[ac],XmNleftPosition,3+x*24); ac++;
    XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
    XtSetArg(al[ac],XmNrightPosition,25+x*24); ac++;
    XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
    XtSetArg(al[ac],XmNtopPosition,67+y*9); ac++;
    XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
    XtSetArg(al[ac],XmNbottomPosition, 75+y*9); ac++;
    command[y]=XmCreatePushButton(form,"label",al,ac);
    XtManageChild(command[y]);
    XtAddCallback(command[y],XmNactivateCallback,
      commandCB,(XtPointer) cmd_number);
  }

  /* Create a separator widget and attach it to the form */

  ac=0;
  XtSetArg(al[ac],XmNtopAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,86); ac++;
  XtSetArg(al[ac],XmNrightAttachment,XmATTACH_FORM); ac++; 
  XtSetArg(al[ac],XmNleftAttachment,XmATTACH_FORM); ac++;  
  XtSetArg(al[ac],XmNbottomAttachment,XmATTACH_NONE); ac++;      
  sep_status=XmCreateSeparator(form,"sep",al,ac);
  XtManageChild(sep_status);   


  /* Message area*/

  /* Create a message widget and attach it to the form */

  ac=0;
  XtSetArg(al[ac],XmNleftAttachment,
    XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,8); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
    XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,92); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
    XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,89); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
    XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,98); ac++;
  XtSetArg(al[ac],XmNeditable,False); ac++;
  XtSetArg(al[ac],XmNeditMode,XmMULTI_LINE_EDIT); ac++;
  message_area=XmCreateScrolledText(form,"message_area",al,ac);
  XtManageChild(message_area);

  /* Create a prompt dialog for target input but leave it unmanaged */

  ac=0;
  XtSetArg(al[ac], XmNselectionLabelString,
    XmStringCreateLtoR("Enter EOD coordinates: ",char_set)); ac++;
  XtSetArg(al[ac], XmNdialogTitle,
    XmStringCreateLtoR("XmTel: target",char_set)); ac++;
  dialog=XmCreatePromptDialog(toplevel,"dialog",al,ac);
  XtAddCallback(dialog, XmNokCallback, dialogCB, (int *) OK);
  XtAddCallback(dialog, XmNcancelCallback, dialogCB, (int *) CANCEL);
  XtUnmanageChild (XmSelectionBoxGetChild (dialog, XmDIALOG_HELP_BUTTON));

  /* Create a dialog to select the log file but leave it unmanaged */

  ac=0;
  XtSetArg(al[ac],XmNmustMatch,True); ac++;
  XtSetArg(al[ac],XmNautoUnmanage,False); ac++;
  select_logfile=XmCreateFileSelectionDialog(toplevel,"select_logfile",al,ac);
  XtAddCallback (select_logfile, XmNokCallback, selectlogCB, (int *) OK);
  XtAddCallback (select_logfile, XmNcancelCallback, selectlogCB, (int *) CANCEL);
  XtUnmanageChild(XmSelectionBoxGetChild(select_logfile,
    XmDIALOG_HELP_BUTTON));
  
  /* Create a dialog to select and read the queue but leave it unmanaged */

  ac=0;
  XtSetArg(al[ac],XmNmustMatch,True); ac++;
  XtSetArg(al[ac],XmNautoUnmanage,False); ac++;
  read_queuefile=XmCreateFileSelectionDialog(toplevel,"read_queuefile",al,ac);
  XtAddCallback (read_queuefile, XmNokCallback, readqueueCB, (int *) OK);
  XtAddCallback (read_queuefile, XmNcancelCallback, readqueueCB, (int *) CANCEL);
  XtUnmanageChild(XmSelectionBoxGetChild(read_queuefile,
    XmDIALOG_HELP_BUTTON));
      
  /* Create the menubar */
  
  create_menus(menu_bar);  
  
  /* Connect to XEphem's goto fifo */

  if (fd_fifo_in > 0)
  {
    XtAppAddInput (context, fd_fifo_in, (XtPointer)XtInputReadMask,
      read_xephem_target, NULL);
  }       

  /* Add the server as input to the user interface */
  
  indi_id = XtAppAddInput (context, indi_socket, (XtPointer)XtInputReadMask,
    indiCB, NULL);
  
  /* Request INDI information */
  
  fprintf (indi_fp, "<getProperties version='%g'/>\n", INDIV);     
  
  /* Create unmanaged control panels */

  setup_model_edit();
  setup_focus();
  setup_rotate();
  setup_fan();
  setup_heater(); 

  /* Start the user interface and telescope communications */
  
  XtRealizeWidget(toplevel);

  XtAppMainLoop(context);

  exit(0);
}

/* Add an item into a menu in the menubar */     

Widget make_menu_item(char *item_name, unsigned long client_data, Widget menu)

{
  int ac;
  Arg al[16];
  Widget item;

  ac = 0;
  XtSetArg(al[ac], XmNlabelString,
      XmStringCreateLtoR(item_name,char_set)); ac++;      
  item=XmCreatePushButton(menu,item_name,al,ac);
  XtManageChild(item);
  XtAddCallback (item,XmNactivateCallback,menuCB,(XtPointer) client_data);      
  XtSetSensitive(item,True);
  return(item);
}

/* Add a toggle item into a menu in the menubar */     

Widget make_menu_toggle(char *item_name, unsigned long client_data, Widget menu)

{
  int ac;
  Arg al[16];
  Widget item;

  ac = 0;
  XtSetArg(al[ac], XmNlabelString,
      XmStringCreateLtoR(item_name,char_set)); ac++;      
  item=XmCreateToggleButton(menu,item_name,al,ac);
  XtManageChild(item);
  XtAddCallback (item,XmNvalueChangedCallback,menuCB,(XtPointer) client_data);      
  XtSetSensitive(item,True);
  return(item);
}

/* Creates a menu on the menu bar */

Widget make_menu(char *menu_name, Widget menu_bar)
 
{
  int ac;
  Arg al[16];
  Widget menu, cascade;

  menu=XmCreatePulldownMenu(menu_bar,menu_name,NULL,0);
  ac=0;
  XtSetArg(al[ac],XmNsubMenuId,menu);  ac++;
  XtSetArg(al[ac],XmNlabelString,
      XmStringCreateLtoR(menu_name,char_set)); ac++;
  cascade=XmCreateCascadeButton(menu_bar,menu_name,al,ac);      
  XtManageChild (cascade); 
  return(menu);
}

/* Create all the menubar menus */

void create_menus(Widget menu_bar)
{
  /* Create the file menu */
  file_menu=make_menu("File",menu_bar);
  newqueue_item=make_menu_item("New Queue",NEWQUEUE,file_menu); 
  newlog_item=make_menu_item("New Log",NEWLOG,file_menu); 
  exit_item=make_menu_item("Exit",EXIT,file_menu);

  /* Create the edit menu */
  edit_menu=make_menu("Edit",menu_bar);
  editqueue_item=make_menu_item("Queue",EDITQUEUE,edit_menu);
  editlog_item=make_menu_item("Log",EDITLOG,edit_menu);
  
  /* Create the pointing menu with toggle options */
  pointing_menu=make_menu("Pointing",menu_bar);
  p_options_toggle_1  = make_menu_toggle("Offset",POINTOPTION1,pointing_menu);
  p_options_toggle_2  = make_menu_toggle("Refraction",POINTOPTION2,pointing_menu);
  p_options_toggle_4  = make_menu_toggle("Polar",POINTOPTION4,pointing_menu);
  p_options_toggle_8  = make_menu_toggle("Model",POINTOPTION8,pointing_menu);

  /* Create the reference pull-down menu */
  ref_menu            = make_menu("Reference",menu_bar);
  ref_target_item     = make_menu_item("Target",REFTARGET,ref_menu);
  ref_wcs_item        = make_menu_item("WCS",REFWCS,ref_menu); 
  ref_clear_item      = make_menu_item("Clear",REFCLEAR,ref_menu);
  ref_save_item       = make_menu_item("Save",REFSAVE,ref_menu);
  ref_recall_item     = make_menu_item("Recall",REFRECALL,ref_menu);
  ref_default_item    = make_menu_item("Default",REFDEFAULT,ref_menu);

  /* Create the model pull-down menu */
  model_menu          = make_menu("Model",menu_bar);
  model_edit_item     = make_menu_item("Edit",MODELEDIT,model_menu);
  model_clear_item    = make_menu_item("Clear",MODELCLEAR,model_menu);
  model_save_item     = make_menu_item("Save",MODELSAVE,model_menu);
  model_recall_item   = make_menu_item("Recall",MODELRECALL,model_menu);
  model_default_item  = make_menu_item("Default",MODELDEFAULT,model_menu);
   
  /* Create the telescope pull-down menu */
  telescope_menu      = make_menu("Telescope",menu_bar);
  focus_item          = make_menu_item("Focus",FOCUS,telescope_menu);
  rotate_item         = make_menu_item("Rotation",ROTATE,telescope_menu);
  fan_item            = make_menu_item("Fan",FAN,telescope_menu);
  heater_item         = make_menu_item("Heater",HEATER,telescope_menu);

  /* Disable menu items that cannot be used with this protocol */
    
  if ( FOCUSDISABLED )
  {
    XtSetSensitive(focus_item,False);
  } 
  if ( ROTATEDISABLED )
  {
    XtSetSensitive(rotate_item,False);
  } 
  if ( THERMALDISABLED )
  {
    XtSetSensitive(fan_item,False);
  } 
  if ( HEATERDISABLED )
  {
    XtSetSensitive(heater_item,False);
  }  
}

/* Callback routines used for menubar */

void menuCB(w,client_data,call_data)
  Widget w;
  int client_data;
  XmAnyCallbackStruct *call_data;
{

  /* If request for a queue file detected, then select a new queue file */

  if (client_data==NEWQUEUE)
  {
    /* Make the queue selection dialog appear */
    
    XtManageChild(read_queuefile);    
  }
  
  /* If request for a new log file detected, then select a new log file */

  if (client_data==NEWLOG)
  {
    /* Make the log selection dialog appear */
    
    XtManageChild(select_logfile);
  } 
  
   
  /* If exit detected, then exit */
  
  if (client_data==EXIT) 
  {
    unlink_fifos();
    exit(0);
  }   

  /* If request to edit the queue is detected, then edit the queue file */

  if (client_data==EDITQUEUE)
  {
    strcpy(buf, editor);
    strcat(buf, " ");
    strcat(buf, queuefile);
    system(buf); 
  } 
  
  /* If request to edit log file detected, then edit the log file */

  if (client_data==EDITLOG)
  {
    strcpy(buf, editor);
    strcat(buf, " ");
    strcat(buf, logfile);
    system(buf); 
  }
  
  /* Detect a change to the toggle for a pointing option */
  /* Check for the state of the option                   */
  /* Set pmodel using bitwise OR or bitwise AND NOT      */
  /* Set the appropriate indiswitch */

  if (client_data==POINTOPTION1)
  {
    if( XmToggleButtonGetState(p_options_toggle_1) )
    {
      pmodel=(pmodel | OFFSET);
      sprintf(telescope_set_prm,"pointing.offset");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);
    }
    else
    {
      pmodel=(pmodel & ~(OFFSET));
      sprintf(telescope_set_prm,"pointing.offset");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, FALSE);
    }  
  } 
      
  if (client_data==POINTOPTION2)
  {
    if( XmToggleButtonGetState(p_options_toggle_2) )
    {
      pmodel=(pmodel | REFRACT);
      sprintf(telescope_set_prm,"pointing.refraction");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);     
    }
    else
    {
      pmodel=(pmodel & ~(REFRACT));
      sprintf(telescope_set_prm,"pointing.refraction");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, FALSE);     
    }
  }         

  if (client_data==POINTOPTION4)
  {
    if( XmToggleButtonGetState(p_options_toggle_4) )
    {
      pmodel=(pmodel | POLAR);
      sprintf(telescope_set_prm,"pointing.polar");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);     
    }
    else
    {
      pmodel=(pmodel & ~(POLAR));
      sprintf(telescope_set_prm,"pointing.polar");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, FALSE);     
    }  
  }         

  if (client_data==POINTOPTION8)
  {
    if( XmToggleButtonGetState(p_options_toggle_8) )
    {
      pmodel=(pmodel | DYNAMIC);
      sprintf(telescope_set_prm,"pointing.dynamic");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE); 
    }
    else
    {
      pmodel=(pmodel & ~(DYNAMIC));
      sprintf(telescope_set_prm,"pointing.dynamic");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, FALSE); 
    }  
  }         

  /* Reference to target request detected */

  if (client_data==REFTARGET)
  {
    fprintf(stderr,"setting reference.target true\n");
    sprintf(telescope_set_prm,"reference.target");
    send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE); 
  }       


  /* Reference to wcs request detected */

  if (client_data==REFWCS)
  {
    sprintf(telescope_set_prm,"reference.wcs");
    send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE); 
  }       

  /* Reference to clear request detected */

  if (client_data==REFCLEAR)
  {
    sprintf(telescope_set_prm,"reference.clear");
    send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE); 
  }       

  /* Reference save request detected */

  if (client_data==REFSAVE)
  {
    sprintf(telescope_set_prm,"reference.save");
    send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE); 
  }       

  /* Reference recall request detected */

  if (client_data==REFRECALL)
  {
    sprintf(telescope_set_prm,"reference.recall");
    send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE); 
  }       

  /* Reference default request detected */

  if (client_data==REFDEFAULT)
  {
    sprintf(telescope_set_prm,"reference.default");
    send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE); 
  }       

  /* If model edit is requested, then handle that here */
  
  if (client_data==MODELEDIT)
  {
    /* Make the model edit panel appear */
    
    XtManageChild(model_edit);    
  }         

  /* If model clear detected, then update offset */

  if (client_data==MODELCLEAR)
  {
    sprintf(telescope_set_prm,"model.clear");
    send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE); 
  } 
  
  /* If model save detected, write to disk  then update display */
  /* Default is set to 0. on startup then reset by the configuration file */

  if (client_data==MODELSAVE)
  {
    sprintf(telescope_set_prm,"model.save");
    send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE); 
  } 
  
  /* If model recall detected, read  then update display */
  /* Default is set to 0. on startup then reset by the configuration file */

  if (client_data==MODELRECALL)
  {
    sprintf(telescope_set_prm,"model.recall");
    send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE); 
  }  

  /* If model default detected, then update offset */
  /* Default is set to 0. on startup then reset by the configuration file */

  if (client_data==MODELDEFAULT)
  {
    sprintf(telescope_set_prm,"model.default");
    send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE); 
  } 

  /* If telescope hardware request detected, then handle that here */
  
  if (client_data==FOCUS)
  {
    /* Make the focus control appear */
    
    XtManageChild(focus_control);    
  }         
  
  if (client_data==ROTATE)
  {
    /* Make the rotate control appear */
    
    XtManageChild(rotate_control);
  
  }         
  
  if (client_data==FAN)
  {
    /* Make fan control appear */
  
    XtManageChild(fan_control);
  }

  if (client_data==HEATER)
  {
    /* Make heater control appear */
  
    XtManageChild(heater_control);
  }

  /* Else noop */   
   
  if (client_data==CANCEL);
}

/* Create the unmanaged mount model edit panel */

void setup_model_edit()
{
  Arg al[10];
  int ac;
  
  ac=0;
  
  XtSetArg(al[ac], XmNselectionLabelString,
    XmStringCreateLtoR("Enter mount model: ",char_set)); ac++;
  XtSetArg(al[ac], XmNdialogTitle,
    XmStringCreateLtoR("Edit mount model",char_set)); ac++;
  model_edit = XmCreatePromptDialog(toplevel,"model_edit",al,ac);
  XtAddCallback(model_edit, XmNokCallback, model_editCB, (XtPointer) OK);
  XtAddCallback(model_edit, XmNcancelCallback, model_editCB, (XtPointer) CANCEL);
  XtUnmanageChild (XmSelectionBoxGetChild (model_edit, XmDIALOG_HELP_BUTTON));
}

/* Callback function for the mount model edit box */
/* Read mount model in units of pixels/hour       */
/* Units conversion done in pointing utility */
     
void model_editCB(w,client_data,call_data)
  Widget w;
  int client_data;
  XmSelectionBoxCallbackStruct  *call_data;
{    
  char *s;
  int testflag;
  double tmpmodelha1, tmpmodeldec1;
  
  switch (client_data)
  {
    case OK:
      XmStringGetLtoR(call_data->value,char_set,&s);
      testflag = sscanf(s, "%lf %lf ", &tmpmodelha1, &tmpmodeldec1);
      if (testflag > 0)
      {
        modelha1  = tmpmodelha1;        
        modeldec1 = tmpmodeldec1;
        sprintf(telescope_set_prm,"model_parameters.ha");
        send_indi_number (indi_fp, telescope_device, telescope_set_prm, modelha1);
        sprintf(telescope_set_prm,"model_parameters.dec");
        send_indi_number (indi_fp, telescope_device, telescope_set_prm, modeldec1);        
      }
            
      XtFree(s);
      break;
    case CANCEL:
       break;
  }
  XtUnmanageChild(w);
}

/* Create the unmanaged focus control panel */

void setup_focus()
{
  Arg al[10];
  int ac;
  unsigned long x;
  EventMask mask;
  
  ac=0;    
  XtSetArg(al[ac],XmNheight,250); ac++;
  XtSetArg(al[ac],XmNwidth,250); ac++;
  XtSetArg(al[ac],XmNautoUnmanage,False); ac++;
  XtSetArg(al[ac],XmNdialogTitle,XmStringCreateLtoR("Focus",char_set)); ac++;
  focus_control=XmCreateFormDialog(toplevel,"Focus",al,ac);
  
  /* Create the focus speed radio box */
  
  ac=0;
  XtSetArg(al[ac],XmNorientation,XmVERTICAL); ac++;
  XtSetArg(al[ac],XmNtopAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,12); ac++; 
  XtSetArg(al[ac],XmNleftAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,5); ac++; 
  XtSetArg(al[ac],XmNrightAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,36); ac++; 
  focus_spd_radio_box=XmCreateRadioBox(focus_control,"focus_spd_radio_box",al,ac);
  XtManageChild(focus_spd_radio_box);
  
  /* Create focus speed toggles in the radio box to hold four buttons */
  
  for (x=0; x<4; x++)
  {
       ac=0;
       XtSetArg(al[ac],XmNlabelString,
           XmStringCreate(focus_spd_name[x],char_set)); ac++;
       focus_spd_toggle[x]=XmCreateToggleButton(focus_spd_radio_box,
           "focus_spd_toggle",al,ac);
       XtManageChild(focus_spd_toggle[x]);
       XtAddCallback (focus_spd_toggle[x], XmNvalueChangedCallback, 
         focus_speedCB, (XtPointer) x);
  }     
 
  /* Set the speed togglebutton on for the initial value of focusspd */

  XmToggleButtonSetState (focus_spd_toggle[3], True, False);
  
  /* Create the push buttons on the panel */
  
  ac=0;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,40); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,40+20); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,24); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,24+20); ac++;
  XtSetArg(al[ac],XmNlabelString,
    XmStringCreateLtoR("In", char_set)); ac++;
  focus_in_button = XmCreatePushButton(focus_control, "focus_in_button",
    al, ac); 
  mask = ButtonReleaseMask | ButtonPressMask;
  XtAddEventHandler(focus_in_button,mask,False,focusCB,(int *) FOCUSIN);         
  XtManageChild(focus_in_button);

  ac=0;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,70); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,70+20); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,24); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,24+20); ac++;
  XtSetArg(al[ac],XmNlabelString,
    XmStringCreateLtoR("Out", char_set)); ac++;
  focus_out_button = XmCreatePushButton(focus_control, "focus_out_button",
    al, ac);
  mask = ButtonReleaseMask | ButtonPressMask;
  XtAddEventHandler(focus_out_button,mask,False,focusCB,(int *) FOCUSOUT);
  XtManageChild(focus_out_button);

  ac=0;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,55); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,55+20); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,55); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,55+10); ac++;
  XtSetArg(al[ac],XmNlabelString,
    XmStringCreateLtoR("000000", char_set)); ac++;
  focus_count_label = XmCreateLabel(focus_control,"label", al, ac); 
  XtManageChild(focus_count_label);

  ac=0;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,10); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,10+20); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,72); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,72+10); ac++;
  XtSetArg(al[ac],XmNlabelString,
    XmStringCreateLtoR("OK", char_set)); ac++;
  focus_ok_button = XmCreatePushButton(focus_control,"focus_ok_button",
    al, ac); 
  mask = ButtonReleaseMask | ButtonPressMask;      
  XtAddEventHandler(focus_ok_button,mask,False, focusCB,(int *) FOCUSOK);
  XtManageChild(focus_ok_button);
}

/* Create the unmanaged rotate control panel */

void setup_rotate()
{
  Arg al[10];
  int ac;
  unsigned long x;
  EventMask mask;
  
  ac=0;    
  XtSetArg(al[ac],XmNheight,250); ac++;
  XtSetArg(al[ac],XmNwidth,250); ac++;
  XtSetArg(al[ac],XmNautoUnmanage,False); ac++;
  XtSetArg(al[ac],XmNdialogTitle,XmStringCreateLtoR("Rotate",char_set)); ac++;
  rotate_control=XmCreateFormDialog(toplevel,"Rotate",al,ac);
  
  /* Create the rotate speed radio box */
  
  ac=0;
  XtSetArg(al[ac],XmNorientation,XmVERTICAL); ac++;
  XtSetArg(al[ac],XmNtopAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,12); ac++; 
  XtSetArg(al[ac],XmNleftAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,5); ac++; 
  XtSetArg(al[ac],XmNrightAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,36); ac++; 
  rotate_spd_radio_box=XmCreateRadioBox(rotate_control,"rotate_spd_radio_box",al,ac);
  XtManageChild(rotate_spd_radio_box);
  
  /* Create rotate speed toggles in the radio box to hold 3 buttons */
  
  for (x=0; x<3; x++)
  {
       ac=0;
       XtSetArg(al[ac],XmNlabelString,
           XmStringCreate(rotate_spd_name[x],char_set)); ac++;
       rotate_spd_toggle[x]=XmCreateToggleButton(rotate_spd_radio_box,
           "rotate_spd_toggle",al,ac);
       XtManageChild(rotate_spd_toggle[x]);
       XtAddCallback (rotate_spd_toggle[x], XmNvalueChangedCallback, 
         rotate_speedCB, (XtPointer) x);
  }     
 
  /* Set the speed togglebutton to initial value of rotatespd */

  XmToggleButtonSetState (rotate_spd_toggle[1], True, False);
  
  /* Create the rotate command radio box */
  
  ac=0;
  XtSetArg(al[ac],XmNorientation,XmVERTICAL); ac++;
  XtSetArg(al[ac],XmNtopAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,12); ac++; 
  XtSetArg(al[ac],XmNleftAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,55); ac++; 
  XtSetArg(al[ac],XmNrightAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,86); ac++; 
  rotate_cmd_radio_box=XmCreateRadioBox(rotate_control,"rotate_cmd_radio_box",al,ac);

  XtManageChild(rotate_cmd_radio_box);
 
  /* Create rotate command toggles in the radio box to hold 3 buttons */
  
  for (x=0; x<3; x++)
  {
       ac=0;
       XtSetArg(al[ac],XmNlabelString,
           XmStringCreate(rotate_cmd_name[x],char_set)); ac++;
       rotate_cmd_toggle[x]=XmCreateToggleButton(rotate_cmd_radio_box,
           "rotate_cmd_toggle",al,ac);
       XtManageChild(rotate_cmd_toggle[x]);
       XtAddCallback (rotate_cmd_toggle[x], XmNvalueChangedCallback, 
         rotate_commandCB, (XtPointer) x);
  }     
 
  /* Set the command off togglebutton */

  XmToggleButtonSetState (rotate_cmd_toggle[2], True, False);
  
  
  /* Create the rotate display and the ok push button on the panel */
  
  ac=0;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,55); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,55+20); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,55); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,55+10); ac++;
  XtSetArg(al[ac],XmNlabelString,
    XmStringCreateLtoR("000.00", char_set)); ac++;
  rotate_count_label = XmCreateLabel(rotate_control,"label", al, ac); 
  XtManageChild(rotate_count_label);

  ac=0;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,10); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,10+20); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,72); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,72+10); ac++;
  XtSetArg(al[ac],XmNlabelString,
    XmStringCreateLtoR("OK", char_set)); ac++;
  rotate_ok_button = XmCreatePushButton(rotate_control,"rotate_ok_button",
    al, ac); 
  mask = ButtonReleaseMask | ButtonPressMask;      
  XtAddEventHandler(rotate_ok_button,mask,False, rotateCB,(int *) ROTATEOK);
  XtManageChild(rotate_ok_button);
}



/* Create the unmanaged fan control panel */

void setup_fan()
{
  Arg al[10];
  int ac;
  unsigned long x;
  EventMask mask;
  
  ac=0;    
  XtSetArg(al[ac],XmNheight,250); ac++;
  XtSetArg(al[ac],XmNwidth,250); ac++;
  XtSetArg(al[ac],XmNautoUnmanage,False); ac++;
  XtSetArg(al[ac],XmNdialogTitle,XmStringCreateLtoR("Fan",char_set)); ac++;
  fan_control=XmCreateFormDialog(toplevel,"fan",al,ac);
  
  /* Create the fan speed radio box */
  
  ac=0;
  XtSetArg(al[ac],XmNorientation,XmVERTICAL); ac++;
  XtSetArg(al[ac],XmNtopAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,12); ac++; 
  XtSetArg(al[ac],XmNleftAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,5); ac++; 
  XtSetArg(al[ac],XmNrightAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,36); ac++; 
  fan_spd_radio_box=XmCreateRadioBox(fan_control,"fan_spd_radio_box",al,ac);
  XtManageChild(fan_spd_radio_box);
  
  /* Create fan speed toggles in the radio box */
  
  for (x=0; x<3; x++)
  {
       ac=0;
       XtSetArg(al[ac],XmNlabelString,
           XmStringCreate(fan_spd_name[x],char_set)); ac++;
       fan_spd_toggle[x]=XmCreateToggleButton(fan_spd_radio_box,
           "fan_spd_toggle",al,ac);
       XtManageChild(fan_spd_toggle[x]);
       XtAddCallback (fan_spd_toggle[x], XmNvalueChangedCallback, 
         fan_speedCB, (XtPointer) x);
  }     
 
  /* Set the speed togglebutton to initial value of fan speed */

  XmToggleButtonSetState (fan_spd_toggle[2], True, False);
  
  ac=0;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,10); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,10+20); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,72); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,72+10); ac++;
  XtSetArg(al[ac],XmNlabelString,
    XmStringCreateLtoR("OK", char_set)); ac++;
  fan_ok_button = XmCreatePushButton(fan_control,"fan_ok_button",
    al, ac); 
  mask = ButtonReleaseMask | ButtonPressMask;      
  XtAddEventHandler(fan_ok_button,mask,False, fanCB,(XtPointer) FANOK);
  XtManageChild(fan_ok_button);
}


/* Create the unmanaged heater control panel */

void setup_heater()
{
  Arg al[10];
  int ac;
  unsigned long x;
  EventMask mask;
  
  ac=0;    
  XtSetArg(al[ac],XmNheight,250); ac++;
  XtSetArg(al[ac],XmNwidth,250); ac++;
  XtSetArg(al[ac],XmNautoUnmanage,False); ac++;
  XtSetArg(al[ac],XmNdialogTitle,XmStringCreateLtoR("Heater",char_set)); ac++;
  heater_control=XmCreateFormDialog(toplevel,"Heater",al,ac);
  
  /* Create the heater power radio box */
  
  ac=0;
  XtSetArg(al[ac],XmNorientation,XmVERTICAL); ac++;
  XtSetArg(al[ac],XmNtopAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,12); ac++; 
  XtSetArg(al[ac],XmNleftAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,5); ac++; 
  XtSetArg(al[ac],XmNrightAttachment,XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,36); ac++; 
  heater_pwr_radio_box=XmCreateRadioBox(heater_control,"heater_pwr_radio_box",al,ac);
  XtManageChild(heater_pwr_radio_box);
  
  /* Create heater power toggles in the radio box */
  
  for (x=0; x<3; x++)
  {
       ac=0;
       XtSetArg(al[ac],XmNlabelString,
           XmStringCreate(heater_pwr_name[x],char_set)); ac++;
       heater_pwr_toggle[x]=XmCreateToggleButton(heater_pwr_radio_box,
           "heater_pwr_toggle",al,ac);
       XtManageChild(heater_pwr_toggle[x]);
       XtAddCallback (heater_pwr_toggle[x], XmNvalueChangedCallback, 
         heater_powerCB, (XtPointer) x);
  }     
 
  /* Set the power togglebutton to initial value of heaterpwr */

  XmToggleButtonSetState (heater_pwr_toggle[2], True, False);
  
  ac=0;
  XtSetArg(al[ac],XmNleftAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNleftPosition,10); ac++;
  XtSetArg(al[ac],XmNrightAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNrightPosition,10+20); ac++;
  XtSetArg(al[ac],XmNtopAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNtopPosition,72); ac++;
  XtSetArg(al[ac],XmNbottomAttachment,
      XmATTACH_POSITION); ac++;
  XtSetArg(al[ac],XmNbottomPosition,72+10); ac++;
  XtSetArg(al[ac],XmNlabelString,
    XmStringCreateLtoR("OK", char_set)); ac++;
  heater_ok_button = XmCreatePushButton(heater_control,"heater_ok_button",
    al, ac); 
  mask = ButtonReleaseMask | ButtonPressMask;      
  XtAddEventHandler(heater_ok_button,mask,False, heaterCB,(XtPointer) HEATEROK);
  XtManageChild(heater_ok_button);
}

      
/* Callback function for the target input box */

void dialogCB(w,client_data,call_data)
  Widget w;
  int client_data;
  XmSelectionBoxCallbackStruct  *call_data;
{    
  char *s;
  int testflag;

  switch (client_data)
  {
    case OK:
      XmStringGetLtoR(call_data->value,char_set,&s);
      switch (target)
      {  
        case RA: 
        testflag=dmstod(s,&targetra);
        sprintf(telescope_set_prm,"target.ra");
        send_indi_number (indi_fp, telescope_device, telescope_set_prm, targetra);
        show_target_coordinates();          
        break;
        
        case DEC: 
        testflag=dmstod(s,&targetdec);
        sprintf(telescope_set_prm,"target.dec");
        send_indi_number (indi_fp, telescope_device, telescope_set_prm, targetdec);
        show_target_coordinates();
        break;
      }
      XtFree(s);
      break;
    case CANCEL:
       break;
  }
  XtUnmanageChild(w);
}

/* Callback function for the speed toggle buttons */

void speedCB(w,client_data,call_data)
  Widget w;
  int client_data;
  XmAnyCallbackStruct *call_data;
{
  int n;

  /* Identify the button code*/
  n=client_data;
     
  /* Assign a speed code */
  /* Indiswitch for slew_speed is type one of many */
  /* Toggle the appropriate switch */
  
   switch (n) {
  
     /* Slew */
     case 0: 
       telspd=SLEW;
       sprintf(telescope_set_prm,"slew_speed.slew");
       send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE); 
       break; 
        
     /* Find */
     case 1: 
       telspd=FIND;
       sprintf(telescope_set_prm,"slew_speed.find");
       send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE); 
       break;
         
     /* Center */
     case 2: 
       telspd=CENTER;
       sprintf(telescope_set_prm,"slew_speed.center");
       send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE); 
       break; 
      
     /* Guide */
     case 3: 
       telspd=GUIDE;
       sprintf(telescope_set_prm,"slew_speed.guide");
       send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE); 
       break;  
   } 
}

/* Callback function for the grid buttons */
/* Called when one of the grid buttons is pressed or released*/
/* Grid buttons control NSEW motion on operator command */
/* Toggle corresponding indiswitch */

void buttonCB(w,client_data,call_event)
  Widget w;
  int client_data;
  XEvent *call_event;
{
  int n;
  int et = call_event->type;
  int dn,up;
   
  /* identify the button number n = 3*x+y */
  n=client_data;

  /* identify the action */
  
  dn = et == ButtonPress;
  up = et == ButtonRelease;
      
  /* respond to the request */

  if(dn)
  {

    switch (n) 
    {
   
      /* NoOp */
      case 0:
      break;

      /* E */
      case 1: 
      sprintf(telescope_set_prm,"slew.east");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);       
      strcpy(message,"Moving east\n");
      show_message();
      break;

      /* NoOp */
      case 2: 
      break;

      /* N */
      case 3: 
      sprintf(telescope_set_prm,"slew.north");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);       
      strcpy(message,"Moving north\n");
      show_message();
      break;

      /* NoOp */
      case 4:
      break;

      /* S */
      case 5: 
      sprintf(telescope_set_prm,"slew.south");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);       
      strcpy(message,"Moving south\n");
      show_message();
      break;

      /* NoOp */
      case 6:
      break;

      /* W */
      case 7:
      sprintf(telescope_set_prm,"slew.west");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);       
      show_message();
      break;

      /* NoOp */
      case 8: 
      break;    
    }
  }
  
  if(up)
  {
    
    switch (n) 
    {
   
      /* NoOp */
      case 0:
      break;

      /* E */
      case 1: 
      sprintf(telescope_set_prm,"slew.east");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, FALSE);       
      strcpy(message,"Moving east\n");
      show_message();
      break;

      /* NoOp */
      case 2: 
      break;

      /* N */
      case 3: 
      sprintf(telescope_set_prm,"slew.north");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, FALSE);       
      strcpy(message,"Moving north\n");
      show_message();
      break;

      /* NoOp */
      case 4:
      break;

      /* S */
      case 5: 
      sprintf(telescope_set_prm,"slew.south");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, FALSE);       
      strcpy(message,"Moving south\n");
      show_message();
      break;

      /* NoOp */
      case 6:
      break;

      /* W */
      case 7:
      sprintf(telescope_set_prm,"slew.west");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, FALSE);       
      show_message();
      break;

      /* NoOp */
      case 8: 
      break;    
    }

    /* Update the console message */
    if (guidecmd == TRUE)
    {
      strcpy(message,"Guiding\n");
      show_message();    
    }
    else
    {
      strcpy(message,"Tracking\n");
      show_message();
    }  
   }
}


/* Callback function for the 6 telescope data command buttons */
/* All of them respond with a mark on XEphem */
/* Epoch switch toggles the epoch of the displayed coordinates */

void teldataCB(w,client_data,call_data)
  Widget w;
  int client_data;
  XmAnyCallbackStruct *call_data;
{
   
  int n, cmd_number;

  /* Identify the button */
  cmd_number=client_data;

  /* Respond to the request */
  n=cmd_number;

  switch (n) 
  {

    /* Update mark on XEphem */
      case 0: 
      mark_xephem_telescope();
      break;

    /* Mark target on XEphem */
    case 1: 
      mark_xephem_target();
      break;

    /* Update mark on XEphem */
    case 2: 
      mark_xephem_telescope();
      break;

    /* Update mark on XEphem */
    case 3: 
      mark_xephem_telescope();
      break;
      
    /* Update telescope display epoch */
    case 4: 
      display_telepoch = display_telepoch + 1;
      display_telepoch = display_telepoch % 2;
      show_telescope_coordinates();
      mark_xephem_telescope();
      break;        
      
    /* Enter target RA from the console */
    case 5: 
      new_target_ra();
      mark_xephem_target(); 
      break;

    /* Enter target Dec from the console */
    case 6: 
      new_target_dec();
      mark_xephem_target();
      break;

    /* Update target display epoch */
    case 7: 
      display_targetepoch = display_targetepoch + 1;
      display_targetepoch = display_targetepoch % 2;
      show_target_coordinates();
      break;
  }
}

/* Callback function for the 8 telescope command buttons */

void commandCB(w,client_data,call_data)
  Widget w;
  int client_data;
  XmAnyCallbackStruct *call_data;
{
  int n, cmd_number;

  /* identify the button */
  cmd_number=client_data;

  /* respond to the request */
  n=cmd_number;

  switch (n) 
  {

    /* Slew to target */
    case 0: 
      /* If guiding turn this feature off first */
      
      if(guidecmd == TRUE)
      {
        guidecmd = FALSE;
        sprintf(telescope_set_prm,"guide.on");
        send_indi_switch (indi_fp, telescope_device, telescope_set_prm, FALSE); 
      }    
      sprintf(telescope_set_prm,"goto.target");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);
      /* Response of INDI server will prompt a status message */ 
      break;     

    /* Save telescope coordinates to the log */
    case 1: 
      save_coordinates();
      break;

    /* Tracking on */
    case 2: 
      sprintf(telescope_set_prm,"track.on");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE); 
      sprintf(telescope_set_prm,"guide.on");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, FALSE);
      trackcmd = TRUE;
      gotocmd = FALSE;
      guidecmd = FALSE;
      strcpy(message,"Tracking on and guiding off\n");
      show_message();
      break; 
 
    /* Recall a target from the log */
    case 3: 
      recall_coordinates();
      break;

    /* Precision closed loop guide */
    case 4: 
      if(trackcmd == FALSE)
      {
        trackcmd = TRUE;
        sprintf(telescope_set_prm,"track.on");
        send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE); 
      }
      guidecmd = TRUE;
      sprintf(telescope_set_prm,"guide.on");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);
      strcpy(message,"Tracking on and guiding on\n");
      show_message();
      break;        

    /* Target meridian at this Dec */
    case 5:       
      sprintf(telescope_set_prm,"set_target.meridian");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);
      targetcmd = TARGETCMDMERIDIAN;
      break;      

    /* Stop all motion */
    case 6: 
      sprintf(telescope_set_prm,"track.on");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, FALSE);       
      trackcmd = FALSE;
      gotocmd = FALSE;
      guidecmd = FALSE;
      strcpy(message,"Tracking and guiding off\n");      
      show_message();
      break; 

    /* Target park position */
    case 7: 
      sprintf(telescope_set_prm,"set_target.park");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);
      targetcmd = TARGETCMDPARK;
      break;        
  }
}

/* Callback function for selecting the log file name */

void selectlogCB(w,client_data,call_data)
  Widget w;
  int client_data;
  XmAnyCallbackStruct *call_data;
 
{
  XmFileSelectionBoxCallbackStruct *s =
      (XmFileSelectionBoxCallbackStruct *) call_data;
  
  /* Do nothing if cancel is selected. */
  if (client_data==CANCEL) 
  {
      XtUnmanageChild(select_logfile);
      return;
  }

  /* Manage memory for file name */
  
  if(logfile != NULL)
  {
    XtFree(logfile);
    logfile = NULL;
  }
  
  /* Get the filename from the file selection box */
  XmStringGetLtoR(s->value, char_set, &logfile);
  XtUnmanageChild(select_logfile);
} 
 

/* Configuration callback only allowed in "tel" or "xmtel1" */


/* Callback function for selecting and reading the queue file  */
 
void readqueueCB(w,client_data,call_data)
  Widget w;
  int client_data;
  XmAnyCallbackStruct *call_data;
{
  XmFileSelectionBoxCallbackStruct *s =
      (XmFileSelectionBoxCallbackStruct *) call_data;
  
  /* Do nothing if cancel is selected. */
  if (client_data==CANCEL) 
  {
    XtUnmanageChild(read_queuefile);
    return;
  }

  /* Manage memory for file name */
  if(queuefile != NULL)
  {
    XtFree(queuefile);
    queuefile = NULL;
  }
  
  /* Get the filename from the file selection box */
  XmStringGetLtoR(s->value, char_set, &queuefile);

  /* Read the queue into memory */
  read_queue();

  XtUnmanageChild(read_queuefile);

}    

/* Callback function for selecting a queue entry */

void selectqueueCB(w,client_data,call_data)
    Widget w;
    XtPointer client_data;
    XtPointer call_data;
{
  XmListCallbackStruct *cbs = (XmListCallbackStruct *) call_data;

  /* Selection is indexed from 1 in XmList and from 0 in our list */
  
  queuechoice =  cbs->item_position - 1;    
  
  /* Diagnostic of the queue entry EOD coordinates                             */
  /*                                                                           */
  /*  printf(" #%d\n",queuechoice);                                            */
  /*  printf(" %s %s\n",queue[queuechoice].name,queue[queuechoice].desc);      */
  /*  printf(" ra: %lf  dec: %lf  mag: %lf\n\n",                               */
  /*    queue[queuechoice].ra,queue[queuechoice].dec,queue[queuechoice].mag);  */
  
  targetra = queue[queuechoice].ra;
  targetdec = queue[queuechoice].dec;
  sprintf(telescope_set_prm,"target.ra");
  send_indi_number (indi_fp, telescope_device, telescope_set_prm, targetra);
  sprintf(telescope_set_prm,"target.dec");
  send_indi_number (indi_fp, telescope_device, telescope_set_prm, targetdec);
}    

/* Callback function for control of telescope focus  */

void focusCB(w,client_data,call_event)
  Widget w;
  int client_data;
  XEvent *call_event;
{
  int n,dn,up;
  int et = call_event->type;
         
  /* Identify the command */
  
  n=client_data;
  
  /* Identify the focus action */
  
  dn = et == ButtonPress;
  up = et == ButtonRelease;     

  /* Check for button release on ok to return */
  
  if (up & (n == FOCUSOK))
  {
    XtUnmanageChild(focus_control);
    return;   
  }
   
  /* Handle the other buttons */
  
  if(dn)
  {
    switch (n) 
    {
      /* Focus in */
      case FOCUSIN: 
        sprintf(telescope_set_prm,"focus.in");
        send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);     
        focuscmd = FOCUSCMDIN;
        strcpy(message,"Focus moving in\n");
        show_message();
        break;

      /* Focus out  */
      case FOCUSOUT:
        sprintf(telescope_set_prm,"focus.out");
        send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);     
        focuscmd = FOCUSCMDOUT;
        strcpy(message,"Focus moving out\n");
        show_message(); 
        break;
    }
  }
  
  if(up)
  {
    switch (n) 
    {
      /* Focus in */
      case FOCUSIN: 
        sprintf(telescope_set_prm,"focus.off");
        send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);     
        focuscmd = FOCUSCMDOFF;
        strcpy(message,"Focus off\n");
        show_message(); 
        break;

      /* Focus out  */
      case FOCUSOUT:
        sprintf(telescope_set_prm,"focus.off");
        send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);     
        focuscmd = FOCUSCMDOFF;
        strcpy(message,"Focus off\n");
        show_message(); 
        break;
    }     
 
    /* Update the console message */

    strcpy(message,"Focus off\n");
    show_message();
 
    /* Display current focus */
    
    show_focus();
  }
    
}

/* Callback function for the focus speed toggle buttons */

void focus_speedCB(w,client_data,call_data)
  Widget w;
  int client_data;
  XmAnyCallbackStruct *call_data;
{
  int n;

  /* Identify the button code*/
  
  n=client_data;
     
  /* Assign a focus speed code */

  switch (n) 
  {
    
    /* Coarse */
    case 0: 
      sprintf(telescope_set_prm,"focus_speed.8x");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);     
      focusspd = FOCUSSPD4;
      break; 
       
    /* Medium */
    case 1: 
      sprintf(telescope_set_prm,"focus_speed.4x");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);     
      focusspd = FOCUSSPD3;
      break;
    /* Fine */
    case 2: 
      sprintf(telescope_set_prm,"focus_speed.2x");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);     
      focusspd = FOCUSSPD2;
      break; 
       
    /* Precise */
    case 3: 
      sprintf(telescope_set_prm,"focus_speed.1x");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);     
      focusspd = FOCUSSPD1;
      break;
  } 
  
}

/* Callback function for control of telescope rotate  */

void rotateCB(w,client_data,call_event)
  Widget w;
  int client_data;
  XEvent *call_event;
{
  int n,dn,up;
  int et = call_event->type;
         
  /* Identify the command */
  
  n=client_data;
  
  /* Identify the rotate action */
  
  dn = et == ButtonPress;
  up = et == ButtonRelease;     

  /* Check for button release on ok to return */
  
  if (up & (n == ROTATEOK))
  {
    XtUnmanageChild(rotate_control);
    return;   
  }
    
  /* Display current rotate */
    
  show_rotation();
}

/* Callback function for the rotate speed toggle buttons */

void rotate_speedCB(w,client_data,call_data)
  Widget w;
  int client_data;
  XmAnyCallbackStruct *call_data;
{
  int n;

  /* Identify the button code*/
  
  n=client_data;
     
  /* Assign a rotate speed code */

  switch (n) 
  {

    /* Fast */
    case 0: 
      sprintf(telescope_set_prm,"rotate_speed.fast");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);     
      rotatespd = ROTATESPDFAST;
      break; 
       
    /* Slow */
    case 1: 
      sprintf(telescope_set_prm,"rotate_speed.slow");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);     
      rotatespd = ROTATESPDSLOW;
      break;
      
    /* Sidereal */
    case 2: 
      sprintf(telescope_set_prm,"rotate_speed.sidereal");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);     
      rotatespd = ROTATESPDSIDEREAL;
      break;            
  } 
}


/* Callback function for the rotate command toggle buttons */

void rotate_commandCB(w,client_data,call_data)
  Widget w;
  int client_data;
  XmAnyCallbackStruct *call_data;
{
  int n;

  /* Identify the button code*/
  
  n=client_data;
     
  /* Assign a rotate direction code */

  switch (n) 
  {

    /* Clockwise */
    case 0: 
      sprintf(telescope_set_prm,"rotate.cw");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);     
      rotatecmd = ROTATECMDCW;
      break; 
       
    /* Counterclockwise */
    case 1: 
      sprintf(telescope_set_prm,"rotate.ccw");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);     
      rotatecmd = ROTATECMDCCW;
      break;
      
    /* Off */
    case 2: 
      sprintf(telescope_set_prm,"rotate.off");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);     
      rotatecmd = ROTATECMDOFF;
      break;            
  } 
}

/* Callback function for control of telescope fan */

void fanCB(w,client_data,call_event)
  Widget w;
  int client_data;
  XEvent *call_event;
{
  int n,dn,up;
  int et = call_event->type;
         
  /* Identify the command */
  
  n=client_data;
  
  /* Identify the fan action */
  
  dn = et == ButtonPress;
  up = et == ButtonRelease;     

  /* Check for button release on ok to return */
  
  if (up & (n == FANOK))
  {
    XtUnmanageChild(fan_control);
    return;   
  }    
}

/* Callback function for the fan speed toggle buttons */

void fan_speedCB(w,client_data,call_data)
  Widget w;
  int client_data;
  XmAnyCallbackStruct *call_data;
{
  int n;

  /* Identify the button code*/
  
  n=client_data;
     
  /* Assign a fan speed code */

  switch (n) 
  {

    /* High */
    case 0: 
      sprintf(telescope_set_prm,"fan.high");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);  
      fancmd = FANCMDHIGH;
      break; 
       
    /* Low */
    case 1:
      sprintf(telescope_set_prm,"fan.low");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);       
      fancmd = FANCMDLOW;
      break;
      
    /* Off */
    case 2: 
      sprintf(telescope_set_prm,"fan.off");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);     
      fancmd = FANCMDOFF;
      break;            
  } 
}

/* Callback function for control of telescope heater */

void heaterCB(w,client_data,call_event)
  Widget w;
  int client_data;
  XEvent *call_event;
{
  int n,dn,up;
  int et = call_event->type;
         
  /* Identify the command */
  
  n=client_data;
  
  /* Identify the heater action */
  
  dn = et == ButtonPress;
  up = et == ButtonRelease;     

  /* Check for button release on ok to return */
  
  if (up & (n == HEATEROK))
  {
    XtUnmanageChild(heater_control);
    return;   
  }    
}

/* Callback function for the heater speed toggle buttons */

void heater_powerCB(w,client_data,call_data)
  Widget w;
  int client_data;
  XmAnyCallbackStruct *call_data;
{
  int n;

  /* Identify the button code*/
  
  n=client_data;
     
  /* Assign a heater speed code */

  switch (n) 
  {

    /* High */
    case 0:
      sprintf(telescope_set_prm,"heater_power.high");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);        
      heatercmd = HEATERCMDHIGH;
      break; 
       
    /* Low */
    case 1: 
      sprintf(telescope_set_prm,"heater_power.low");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);        
      heatercmd = HEATERCMDLOW;
      break;
      
    /* Off */
    case 2: 
      sprintf(telescope_set_prm,"heater_power.high");
      send_indi_switch (indi_fp, telescope_device, telescope_set_prm, TRUE);        
      heatercmd = HEATERCMDOFF;
      break;            
  } 
}


/* Show latest message on the user interface */

void show_message()
{
  XmString xmstr;
  
  xmstr = XmStringCreateLtoR(message,char_set);
  XmTextSetString(message_area, message); 
  XmStringFree(xmstr);
} 

/* Show telescope coordinates for telepoch in the button */

void show_telescope_coordinates()
{
  double tmpra, tmpdec;
  XmString xmstr;
  Arg al[10];
  int ac;

  switch (display_telepoch)
  {
          
    case J2000:
    
      /* Display Epoch 2000.0 coordinates */      
    
      xmstr = XmStringCreateLtoR("2000.0",char_set);
      ac=0; 
      XtSetArg(al[ac],XmNlabelString,xmstr); ac++; 
      XtSetValues(get_telepoch,al,ac);            
      tmpra = telra2;
      tmpdec = teldec2;
      
      break;
    
    case  EOD:
    
      /* Display EOD coordinates */      
    
      xmstr = XmStringCreateLtoR("Now",char_set);
      ac=0; 
      XtSetArg(al[ac],XmNlabelString,xmstr); ac++; 
      XtSetValues(get_telepoch,al,ac);
      tmpra=telra;
      tmpdec=teldec; 
      break;
  }   

  dtodms(buf, &tmpra);
  xmstr = XmStringCreateLtoR(buf,char_set);
  ac=0;
  XtSetArg(al[ac],XmNlabelString,xmstr); ac++;
  XtSetValues(get_telra,al,ac);
  
  dtodms(buf, &tmpdec);
  xmstr = XmStringCreateLtoR(buf,char_set);
  ac=0;
  XtSetArg(al[ac],XmNlabelString,xmstr); ac++;
  XtSetValues(get_teldec,al,ac);    

  XmStringFree(xmstr);
  
}  
  
/* Show target coordinates for targetepoch in the button */

void show_target_coordinates()
{
  double tmpra, tmpdec;
  XmString xmstr;
  Arg al[10];
  int ac;

  switch (display_targetepoch)
  {
    
    case J2000:
    
      /* Display epoch 2000.0 target coordinates */
    
      ac=0;      
      xmstr = XmStringCreateLtoR("2000.0",char_set);
      XtSetArg(al[ac],XmNlabelString,xmstr); ac++; 
      XtSetValues(get_targetepoch,al,ac);
      tmpra=targetra2;
      tmpdec=targetdec2;        
      break;
    
    case  EOD:
    
      /* Display EOD target coordinates */
    
      xmstr = XmStringCreateLtoR("Now",char_set);
      ac=0; 
      XtSetArg(al[ac],XmNlabelString,xmstr); ac++; 
      XtSetValues(get_targetepoch,al,ac);
      tmpra=targetra;
      tmpdec=targetdec;          
      break;
  }      
 
  dtodms(buf, &tmpra);
  xmstr = XmStringCreateLtoR(buf,char_set);
  ac=0;
  XtSetArg(al[ac],XmNlabelString,xmstr); ac++;
  XtSetValues(get_targetra,al,ac);   

  dtodms(buf, &tmpdec);
  xmstr = XmStringCreateLtoR(buf,char_set);
  ac=0;
  XtSetArg(al[ac],XmNlabelString,xmstr); ac++; 
  XtSetValues(get_targetdec,al,ac);       

  XmStringFree(xmstr);
}  

/* Export current telescope coordinates to XEphem */

void mark_xephem_telescope()     
{
  double tmpra, tmpdec;
  double rarad,decrad;
  char outbuf[101];
  
  /* Recall the telescope's 2000.0 coordinates */
  
  tmpra=telra2;
  tmpdec=teldec2;         

  /* Convert to radians */
  
  rarad=tmpra*PI/12.;  
  decrad=tmpdec*PI/180.;  
  
  /* Send the values to the fifo */
  
  snprintf(outbuf,sizeof(outbuf),"RA:%9.6f Dec:%9.6f Epoch:2000.000\n",
    rarad,decrad);  
  write(fd_fifo_out, outbuf, strlen(outbuf));
}

/* Export current target coordinates to XEphem */

void mark_xephem_target()     
{
  double tmpra, tmpdec;
  double rarad,decrad;
  char outbuf[101]; 
  
  /* Recall the telescope's 2000.0 coordinates */
  
  tmpra=targetra2;
  tmpdec=targetdec2;         
  
  /* Convert to radians */
  
  rarad=tmpra*PI/12.;  
  decrad=tmpdec*PI/180.; 
  
  /* Send the values to the fifo */  
  
  snprintf(outbuf,sizeof(outbuf),"RA:%9.6f Dec:%9.6f Epoch:2000.000\n",
    rarad,decrad);  
  write(fd_fifo_out, outbuf, strlen(outbuf));
}

/* Read XEphem goto fifo and parse target coordinates */

void read_xephem_target() 
{
  int i,j,k;
  double rahr,ramin,rasec;
  double decdeg,decmin,decsec;
  double tmpra,tmpdec,tmpepoch;
  double tmp_pm_ra=0.;
  double tmp_pm_dec=0.;
  double tmpmag;
  double tmpdat;
  char fifostr[121];
  char rastr[121];
  char *raptr=rastr;  
  char inbuf[2];
  
  /* A typical fixed target string is:                            */

  /* Regulus,f|M|B7,10:08:22.3, 11:58:02,  1.35,2000,0            */
  /* Name, type info, RA | pm, Dec | pm, Mag, Epoch, Identifier   */

  /* Return if the buffer is not open                             */

  if(fd_fifo_in<0)
    return;

  /* Read the buffer */

  k=0;
  while(1) {
    i = read(fd_fifo_in,inbuf,1);
    if(i==-1) {
      fifostr[k]='\0';
      break;
    }
    if(k<120)
    fifostr[k++]=inbuf[0];
  }
  
  /* Diagnostic of fifo string contents */
  
  /*  printf("Fifo in: %s\n",fifostr); */
    
  /* Examples from XEphem version 3.6.xx -- */
  
  /* Arcturus,f|V|K1,14:15:39.7|-1093, 19:10:57|-1998,-0.04,2000,0  */
  /* HD 124953,f|V|A8,14:16:04.2|43, 18:54:43|-28,5.98,2000,0       */
  /* Crt Delta-12,f|S|G8,11:19:20.5|-122,-14:46:43|208,3.56,2000,0  */
  /* GSC 0838-0788,f|S,10:32:32.4,  9:06:17,13.84,2000,0            */
  /* M67,f|O|T2, 8:51:24.0, 11:49:00,6.90,2000,1500|1500|0          */
  /* NGC 2539,f|O|T2, 8:10:36.9,-12:49:14,6.50,2000,900|900|0       */
  /* Venus,P                                                        */


  /* If the buffer was not empty then extract its contents */

  if (k>1) 
  {
    fifostr[k]='\0';

    /* Remove any white space */

    i=0;
    j=0;
    while (i<=k)
    {
      if( fifostr[i]!=' ' )
      {
        rastr[j]=fifostr[i];
        j++;
      }
      i++;
    }  
    
    /* Diagnostic of clean coordinate string contents */

    /*  printf("Coordinate str: %s\n",rastr); */
      

    /* Skip the name and description fields by searching for delimiting commas */
    /* This will skip orbital objects for which XEphem does not pass RA and Dec */   

    raptr=rastr;
    i=0;
    while(i<2)  
    {
      raptr=strchr(raptr,',');
      if(raptr==NULL) break;
      raptr++;
      i++;
    }
    
    
    /* Parsing will have proceeded to the RA field if this is a catalog object */

    if(i==2)
    { 

      /* Scan and save RA data */ 
      /* Note that the XEphem proper motion is in milliarcseconds per year */ 
      /* Our pm_ra is in seconds of time per year */

      tmpdat=0.;
      j=sscanf(raptr,"%lf:%lf:%lf|%lf", 
        &rahr,&ramin,&rasec,&tmpdat);

      if(j>0) 
      {
        tmpra = rahr + ramin/60. + rasec/3600.;
        tmp_pm_ra=tmpdat/15000.;
      }
       
      /* Try to advance pointer to the Dec field */

      raptr=strchr(raptr,',');
      if(raptr != NULL)
      {
       raptr++;
       i++;
      }
    }
     

    if(i==3)
    {  

      /* Now pointing at the Dec field so scan and save Dec data */
      /* Note that the XEphem proper motion is in milliarcseconds per year */ 
      /* Our pm_dec is in seconds of arc per year */

      tmpdat=0.;
      j=sscanf(raptr,"%lf:%lf:%lf|%lf", 
        &decdeg,&decmin,&decsec,&tmpdat);

      if(j>0) 
      {
        if(decdeg>0.)
        {
          tmpdec = decdeg + decmin/60. + decsec/3600.;
        }
        else
        { 
          tmpdec = decdeg - decmin/60. - decsec/3600.;
        }
        tmp_pm_dec=tmpdat/1000.;
      }     

      /* Try to advance pointer to the Magnitude field */

      raptr=strchr(raptr,',');
      if(raptr != NULL)
      {
       raptr++;
       i++;
      }
            
    }

    if(i==4)
    {  

      /* Now pointing at the Magnitude field so scan and save magnitude data */ 

      j=sscanf(raptr,"%lf", &tmpdat);

      if(j>0) 
      {
        tmpmag=tmpdat;
      }
    
      /* Try to advance pointer to the Epoch field */

      raptr=strchr(raptr,',');
      if(raptr != NULL)
      {
       raptr++;
       i++;
      }   
    }
  
    if(i==5)
    {  

      /* Now pointing at the Epoch field so scan and show Epoch data */ 

      j=sscanf(raptr,"%lf", &tmpepoch);

      if(j>0) 
      {

         /* The catalog entry epoch has been found */
         /* Apply proper motion to EOD */

         ProperMotion(tmpepoch, &tmpra, &tmpdec, tmp_pm_ra, tmp_pm_dec);        
      
         /* Find the apparent coordinates including */
         /*   precession, nutation, and stellar aberration for the EOD */

         if ((tmpepoch > 1999.99) && (tmpepoch < 2000.01)) 
         {
           
           /* Epoch 2000.0 coordinates apply */
           /* Find the apparent position of the object for EOD */
           
           targetra = tmpra;
           targetdec = tmpdec;
           Apparent(&targetra, &targetdec, 1);
         }
         else 
         {
         
           /* Some other epoch applies.  Caution!!  */
           /* Set the epoch to 2000.0 and then proceed as before. */
           /* If coordinates are not FK5 this will not be accurate. */
                     
           PrecessToEOD(tmpepoch, &tmpra, &tmpdec);
           PrecessToEpoch(2000.0, &tmpra, &tmpdec);
          
           /* Epoch 2000.0 coordinates now apply */
           /* Find the apparent position of the object for EOD */

           targetra = tmpra;
           targetdec = tmpdec;
           Apparent(&targetra, &targetdec, 1);
         }
                           
         /* Send the selection to INDI */
         
         sprintf(telescope_set_prm,"target.ra");
         send_indi_number (indi_fp, telescope_device, telescope_set_prm, targetra);
         sprintf(telescope_set_prm,"target.dec");
         send_indi_number (indi_fp, telescope_device, telescope_set_prm, targetdec);         
         
         /* Show the entry at the epoch selected for display */         
         
         show_target_coordinates();
      }
      else
      {
         
         /* No epoch was found so assume EOD coordinates */
         
         targetra = tmpra;
         targetdec = tmpdec;     
         
         /* Send the selection to INDI */
         
         sprintf(telescope_set_prm,"target.ra");
         send_indi_number (indi_fp, telescope_device, telescope_set_prm, targetra);
         sprintf(telescope_set_prm,"target.dec");
         send_indi_number (indi_fp, telescope_device, telescope_set_prm, targetdec);         
                     
         /* Show the entry at the epoch selected for display */
         
         show_target_coordinates();  
      }     
    }  
  }
}

/* Write current telescope coordinates to the xmtel log file and to history */

void save_coordinates()           
{
  char decstr[20], rastr[20];
  time_t timebuf;
  char *gmtstr;
  
  time(&timebuf);
  gmtstr = asctime(gmtime(&timebuf));
  if (gmtstr[strlen(gmtstr) - 1] == '\n') gmtstr[strlen(gmtstr) - 1]= '\0';

  dtodms(rastr,&telra);
  dtodms(decstr,&teldec);
  fp_log=fopen(logfile, "a");
  fprintf(fp_log,"%s  %s  %s\n",gmtstr,rastr,decstr);
  fflush(fp_log);
  fclose(fp_log);
  
  /* Expand the buffer up to a maximum of 100 entries */
  
  if (nhistory < 100) 
  {
    nhistory++;
  }
  
  /* Increment the ring counter */
  
  last_history_saved++;
  
  /* Test for wraparound and reset history counters */
  
  if (last_history_saved > 99) 
  {
    last_history_saved=0;
  }
  last_history_recalled = last_history_saved - 1;
  
  /* Save at the current telescope coordinates in the last_history location */
  
  strncpy(history[last_history_saved].name,gmtstr,79);
  history[last_history_saved].ra = telra;
  history[last_history_saved].dec = teldec;

}

/* Read a new target from the xmtel history file */

void recall_coordinates()           
{
  
  if ( last_history_saved >= 0 )
  {
    last_history_recalled++;
    if (last_history_recalled > nhistory -1 )
    {
      last_history_recalled = 0;
    }
    targetra = history[last_history_recalled].ra;
    targetdec = history[last_history_recalled].dec;

    sprintf(telescope_set_prm,"target.ra");
    send_indi_number (indi_fp, telescope_device, telescope_set_prm, targetra);
    sprintf(telescope_set_prm,"target.dec");
    send_indi_number (indi_fp, telescope_device, telescope_set_prm, targetdec);    
    
    show_target_coordinates();
    strcpy(message,history[last_history_recalled].name);
    strcat(message,"\n");
    show_message();
  }
  else
  {
    strcpy(message,"Nothing yet saved.\n");
    show_message();
  }
}

/* Read the queue line by line */
/* Parse the edb format */

void read_queue()
{
  char queuestr[121];
  char *queueptr=queuestr;
  int i,j,k;
  double rahr,ramin,rasec;
  double decdeg,decmin,decsec;
  double tmpra,tmpdec,tmpepoch;
  double tmp_pm_ra=0.;
  double tmp_pm_dec=0.;
  double tmpmag;
  double tmpdat;

  Arg al[10];
  int ac;
  
  strcpy(message,"Reading the queue ");
  strcat(message,queuefile);
  strcat(message,"\n");
  show_message();
  
  fp_queue=fopen(queuefile, "r");

  /* Queue is indexed from 0 so we start with -1 here */
  
  nqueue = -1;
  while ( queuestr == fgets(queuestr,120,fp_queue) )
  {
    
    /* Diagnostic of coordinate string contents */    
    /* printf("%d %s\n",nqueue + 1, queuestr); */
    /* printf("strlen  output is %d \n",strlen(queuestr)); */
     
    /* If the buffer is not empty and not a comment then extract its parts */

    k=strlen(queuestr);
    if ( (k>1) && ( queuestr[0] != '#' ) ) 
    {

      /* Increment the counter for the number of queue entries */
      
      nqueue++;
        
      /* Prepare the string by overwriting the newline character at the end */
      
      queuestr[k]='\0';
      queueptr=queuestr;
      
      /* The name and description fields may have useful white space */
      
      /* Identify the name field and copy it to the local queue */

      j=strcspn(queueptr,",");
      strncpy(queue[nqueue].name,queueptr,j);
      queueptr = queueptr + j + 1;
      
      /* Identify the description field and copy it to the local queue */
        
      j=strcspn(queueptr,",");
      strncpy(queue[nqueue].desc,queueptr,j);

      /* Remove any white space as the entire string is copied on itself */

      i=0;
      j=0;
      while (i<=k)
      {
        if( queuestr[i]!=' ' )
        {
          queuestr[j]=queuestr[i];
          j++;
        }
        i++;
      }  

      /* Diagnostic of clean coordinate string contents */
      /* printf("Queue clean str: %s\n",queuestr);      */
      

      /* Set the pointer at the beginning of the clean string */

      queueptr=queuestr;
      i=0;
      
      /* Skips the cleaned name and description fields */

        while(i<2)  
        {
          queueptr=strchr(queueptr,',');
          if(queueptr==NULL) break;
          queueptr++;
          i++;
        }
      

      /* Parsing will have proceeded to the RA field if this is an object */

      if(i==2)
      { 

        /* Scan and save RA data in a temporary buffer -- epoch not yet read */ 
        /* Queue proper motion is in milliarcseconds per year */ 
        /* Our pm_ra is in seconds of time per year */

        tmpdat=0.;
        j=sscanf(queueptr,"%lf:%lf:%lf|%lf", 
          &rahr,&ramin,&rasec,&tmpdat);

        if(j>0) 
        {
          tmpra = rahr + ramin/60. + rasec/3600.;
          tmp_pm_ra=tmpdat/15000.;
        }
         
        /* Try to advance pointer to the Dec field */

        queueptr=strchr(queueptr,',');
        if(queueptr != NULL)
        {
          queueptr++;
          i++;
        }
      }
     
      if(i==3)
      {  

        /* Now pointing at the Dec field so scan and save Dec data */
        /* Queue proper motion is in milliarcseconds per year */ 
        /* Our pm_dec is in seconds of arc per year */

        tmpdat=0.;
        j=sscanf(queueptr,"%lf:%lf:%lf|%lf", 
        &decdeg,&decmin,&decsec,&tmpdat);

        if(j>0) 
        {
          if(decdeg>0.)
          {
            tmpdec = decdeg + decmin/60. + decsec/3600.;
          }
          else
          { 
            tmpdec = decdeg - decmin/60. - decsec/3600.;
          }
          tmp_pm_dec=tmpdat/1000.;
        }     

        /* Try to advance pointer to the Magnitude field */

        queueptr=strchr(queueptr,',');
        if(queueptr != NULL)
        {
          queueptr++;
          i++;
        }
            
      }

      if(i==4)
      {  

        /* Scan and save magnitude */ 

        j=sscanf(queueptr,"%lf", &tmpmag);

        if(j>0) 
        {
          queue[nqueue].mag=tmpmag;
        }
    
        /* Try to advance pointer to the Epoch field */

        queueptr=strchr(queueptr,',');
        if(queueptr != NULL)
        {
          queueptr++;
          i++;
        }   
    
      }
  
      if(i==5)
      {  

        /* Scan for Epoch */ 

        j=sscanf(queueptr,"%lf", &tmpepoch);

        if(j>0) 
        {

          /* The queue entry epoch has been found */
          /* Apply proper motion to EOD */

          ProperMotion(tmpepoch, &tmpra, &tmpdec, tmp_pm_ra, tmp_pm_dec);        
      
          /* Find the apparent coordinates including */
          /*   precession, nutation, and stellar aberration for the EOD */

          if ((tmpepoch > 1999.99) && (tmpepoch < 2000.01)) 
          {
           
            /* Epoch 2000.0 coordinates apply */
            /* Find the apparent position of the object for EOD */
           
            Apparent(&tmpra, &tmpdec, 1);

          }
          else 
          {
         
            /* Some other epoch applies.  Caution!!  */
            /* Set the epoch to 2000.0 and then proceed as before. */
            /* If coordinates are not FK5 this will not be accurate. */
                     
            PrecessToEOD(tmpepoch, &tmpra, &tmpdec);
            PrecessToEpoch(2000.0, &tmpra, &tmpdec);
          
            /* Epoch 2000.0 coordinates now apply */
            /* Find the apparent position of the object for EOD */
 
            Apparent(&tmpra, &tmpdec, 1);
          }
                           
          /* Copy the temporary coordinates to the queue entry */         
         
        }
        else
        {
         
           /* No epoch was found so assume EOD coordinates */
           /* Nothing to do */
                  
        }
        
        /* Copy the temporary coordinates to the queue entry */              
      
        queue[nqueue].ra = tmpra;
        queue[nqueue].dec = tmpdec;
      
      }  

      /* Diagnostic of the queue entry                              */
      /*                                                            */
      /* printf(" #%d\n",nqueue);                                   */
      /* printf(" %s %s\n",queue[nqueue].name,queue[nqueue].desc);  */
      /* printf(" ra: %lf  dec: %lf  mag: %lf\n\n",                 */
      /*   queue[nqueue].ra,queue[nqueue].dec,queue[nqueue].mag);   */
 
    }

  } 


  fclose(fp_queue);

  /* Increment nqueue so that it now reads the number of entries to the queue */
  
  nqueue++;

  XmStringTable queue_list =
        (XmStringTable) XtMalloc ( nqueue * sizeof (XmString) ); 
  for (i=0; i < nqueue; i++)
  {
    queue_list[i] = XmStringCreateLocalized (queue[i].name); 
  }
  ac=0; 
  XtSetArg(al[ac],XmNitemCount,nqueue); ac++; 
  XtSetArg(al[ac],XmNitems,queue_list); ac++;
  XtSetValues(queue_area,al,ac);  
  for (i = 0; i < nqueue; i++)
  {
    XmStringFree(queue_list[i]);
  }
  free(queue_list);

}

/* User input of target RA */

void new_target_ra() 
{
  target=RA;
  XtManageChild(dialog);
}

/* User input of target Dec */

void new_target_dec()
{
  target=DEC;
  XtManageChild(dialog);
}

/* Show new model parameters in the message window */

void show_model_parameters(void)
{
  sprintf(message,"Model parameters for HA: %8.3f  Dec: %8.3f\n", modelha1, modeldec1);      
  show_message();
}


/* Show telescope focus in microns relative to startup position */

void show_focus(void)
{
  XmString xmstr;
  Arg al[10];
  int ac;
  sprintf(buf, "%6.1f", telfocus);
  xmstr = XmStringCreateLtoR(buf,char_set);
  ac=0;
  XtSetArg(al[ac],XmNlabelString,xmstr); ac++;
  XtSetValues(focus_count_label,al,ac);
  XmStringFree(xmstr);
}  

/* Show telescope rotator in degrees */

void show_rotation(void)
{
  XmString xmstr;
  Arg al[10];
  int ac;
  sprintf(buf, "%7.2f", telrotation);
  xmstr = XmStringCreateLtoR(buf,char_set);
  ac=0;
  XtSetArg(al[ac],XmNlabelString,xmstr); ac++;
  XtSetValues(rotate_count_label,al,ac);
  XmStringFree(xmstr);
}    
  
/* Show telescope temperature in Celsius in the message window */

void show_temperature(void)
{
  if (THERMALDISABLED == 0)
  {
    sprintf(message,"Primary temperature %2.1f C\n",teltemperature);
    show_message();
  }    
}    
     
/* Start fifo links */

void link_fifos()
{  
  fd_fifo_in = open("/usr/local/xephem/fifos/xephem_loc_fifo", 
    O_RDWR | O_NONBLOCK);


  if (fd_fifo_in<0) {
    printf("Unable to open /usr/local/xephem/fifos/xephem_loc_fifo. \n");
    }
  fd_fifo_out = open("/usr/local/xephem/fifos/xephem_in_fifo", 
    O_RDWR | O_NONBLOCK);


  if (fd_fifo_out<0) {
    printf("Unable to open /usr/local/xephem/fifos/xephem_in_fifo. \n");
    }
}

/* Stop fifo links */

void unlink_fifos()
{
  if (fd_fifo_in!=-1)
    close(fd_fifo_in);
  if (fd_fifo_out!=-1)
    close(fd_fifo_out);  
}

/* Convert string deg:min:sec or hr:min:sec to a double */

int dmstod (char *instr, double *datap)
{
  double h=0., m=0., s=0.;
  int negflag;
  int convertflag;
  
  while (isspace(*instr))
    instr++;
  if (*instr == '-') {
    negflag = 1;
    instr++;
  } else
    negflag = 0;
    
  convertflag = sscanf (instr,  "%lf%*[:]%lf%*[:]%lf", &h, &m, &s);
  if (convertflag < 1)
    return (-1);
  *datap = h + m/60. + s/3600.;
  if (negflag)
    *datap = - *datap;
  return (0);
}
    
/* Convert double to string deg:min:sec or hr:min:sec */
/* This routine has memory leaks that should be fixed */

void dtodms (char *outstr, double *dmsp)
{

  int d=0, m=0, s=0;
  int negflag;
  double dms, ms;
   
  dms = *dmsp;

  /* If negative, treat as positive and set sign in string */

  if (dms < 0 )
  {
    negflag = 1;
    dms = - dms;
  }
  else
  {
    negflag = 0;
  }

  /* Allow for truncation if input is in seconds of arc */
  
  dms = dms + 0.5/3600.;

  /* Get the whole part in degrees or hours, as needed */  

  d = (int) dms;
  
  /* Could test here for overflow at 24h or 360d          */
  /*   but we'll assume that's been done before the call. */
    
  ms = dms - (double) d;
  ms = ms*60.;
  m = (int) ms;
  s = (int) 60.*(ms - (double) m);
  if (negflag)
  {
    sprintf(outstr,"-%02d:%02d:%02d",d,m,s);
  }
  else
  {
    sprintf(outstr,"%02d:%02d:%02d",d,m,s);      
  }
}

/* Indi client utilities */

/* Connect to the server */
/* Return TRUE if connection established and FALSE otherwise */

int connect_indi(void)
{
  struct sockaddr_in serv_addr;
  struct hostent *hp;

  /* Lookup host address */
  hp = gethostbyname (host);

  /* Create a socket to the INDI server */
  (void) memset ((char *)&serv_addr, 0, sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_addr.s_addr =
    ((struct in_addr *)(hp->h_addr_list[0]))->s_addr;
  serv_addr.sin_port = htons(port);
  indi_socket = socket (AF_INET, SOCK_STREAM, 0);
  if (indi_socket < 0)
  {
    fprintf(stderr,"Unable to open server socket\n");
    return (0);
  }

  /* Connect to the server */
  connect (indi_socket,(struct sockaddr *)&serv_addr,sizeof(serv_addr));

  /* Open a file for writing to the server */
  indi_fp = fdopen (indi_socket, "w");
  
  if (indi_fp == NULL)
  {
    fprintf(stderr,"Unable to open server connections\n");
    return (0);    
  }
  
  return (1);
}


/* Disconnect the server */
/* Return server state TRUE or FALSE */

int disconnect_indi(void)
{
  
  /* Close socket and file pointer */
  if (indi_socket >= 0) 
  {
    fclose (indi_fp);     
    indi_fp = NULL;
    indi_socket = -1;
    return (0);
  }
  else
  {
    return (1);
  }    
}

 
/* Implement callback for an incoming INDI command */
/* Adapted from Elwood Downey's INDI code */

void indiCB (XtPointer client, int *fdp, XtInputId *idp)
{
  char ibuf[32];  
  char msg[1024];
  int i, nr;

  /* Read INDI command */

  nr = read (*fdp, ibuf, sizeof(ibuf));
  if (nr <= 0) 
  {
    if (nr < 0)
    {
      fprintf(stderr, "INDI input error\n");
    }
    else
    {
      fprintf(stderr, "INDI closed connection");
    }
      
    disconnect_indi();
    
    /* Clean up lilxml */
    if (lillp) 
    {
      delLilXML (lillp);
      lillp = NULL;
    }
    
    /* Remove input */
    if (indi_id) 
    {
      XtRemoveInput (indi_id);
      indi_id = 0;
    }
            
    return;
  }
     
  /* Process each character */
  
  for (i = 0; i < nr; i++) 
  {
    XMLEle *root = readXMLEle (lillp, (int)ibuf[i], msg);
    if (root) 
    {
            
      if (diagnostics == 1)
      {
        printf("New xml element\n");
        prXMLEle (stdout, root, 0); 
      }

      if (launch_indi_command (root, msg) < 0) 
      {
        fprintf (stderr, "INDI input error: %s\n", msg);
      } 
      delXMLEle (root);
    } 
    else if (msg[0]) 
    {
      fprintf (stderr, "INDI input: %s\n", msg);
    }
  }
}

/* Check incoming INDI tag and launch the command */
/* The commands are tailored specifically for this interface */
/* Returns 0 on success and -1 on error           */
/* Adapted from Elwood Downey's INDI code         */

int launch_indi_command (XMLEle *root, char *errmsg)
{
  static struct {
    char *cmdname;
     int (*fp)(XMLEle *root, char errmsg[]);
  } commands[] = {
      /* Must be sorted by cmdname */
      {"defBLOBVector",   set_indi_blob},
      {"defLightVector",  set_indi_light},
      {"defNumberVector", set_indi_number},
      {"defSwitchVector", set_indi_switch},
      {"defTextVector",   set_indi_text},
      {"message",         set_indi_message},
      {"newNumberVector", set_indi_number},
      {"newSwitchVector", set_indi_switch},
      {"newTextVector",   set_indi_text},
      {"setBLOBVector",   set_indi_blob},
      {"setLightVector",  set_indi_light},
      {"setNumberVector", set_indi_number},
      {"setSwitchVector", set_indi_switch},
      {"setTextVector",   set_indi_text},
  };

  char *tag = tagXMLEle(root);
  int u = XtNumber(commands)-1;
  int l = 0;
  int diff = -1;
  int m = -1;
  
  /* Binary search for known xml tag*/
  while (l <= u) 
  {
    m = (l+u)/2;
    diff = strcmp (tag, commands[m].cmdname);
    if (diff == 0)
    {
      break;
    }
    if (diff < 0)
    {
      u = m-1;
    }
    else
    {
      l = m+1;
    }
  }

  /* Launch the command for a valid tag*/
  if (diff == 0)
  {    
    ((*commands[m].fp) (root, errmsg));
    return (0);
  }
  
  /* An error generates a message and a flag */
  sprintf (errmsg, "Unrecognized INDI tag %s", tag);
  return (-1);
}

/* int send_indi_number (char *dev, char *prm, double value)             */

/* Derived from the command line programs get/setINDI.c by Elwood Downey */

/* Connects to an INDI server                                            */
/* Sends  property.elements=value to a device                            */
/* Returns (0) success,  (1) send error, (2)  trouble opening server     */
/* May fail without notice if the server is not running                  */

/* dev is the device                                                     */
/* prm is the p.e string                                                 */
/* value is the new value                                                */          


int send_indi_number (FILE *fp, char *dev, char *prm, double value)
{
  char  p[1024], e[2048];
  
  /* Parse parameter string to components */
  
  sscanf (prm, "%[^.].%[^.]",  p, e);
  
  /* Create and send an XML string                        */
  
  /* Sample format:                                       */
  /*                                                      */
  /* setINDI  ccd.ImageExposure.Seconds=2                 */
  /*                                                      */
  /* <newNumberVector device='ccd' name='ImageExposure'>  */
  /* <oneNumber name='Seconds'>'2'</oneNumber>            */
  /* </oneNumberVector>                                   */
  
  fprintf (fp, "<newNumberVector device='%s' name='%s'>\n",dev,p);
  fprintf (fp, "  <oneNumber name='%s'>%lf</oneNumber>\n",e,value);
  fprintf (fp, "</newNumberVector>\n");
  fflush (fp);
  if (feof(fp) || ferror(fp)) 
  {
    fprintf (stderr, "Send error\n");
    return 1;
  }
  return 0;
}

/* int send_indi_switch (char *dev, char *prm, int state)                 */

/* Derived from the command line programs get/setINDI.c by Elwood Downey */

/* Connects to an INDI server                                            */
/* Sends  property.elements=value to a device                            */
/* Returns (0) success,  (1) send error, (2)  trouble opening server     */
/* May fail without notice if the server is not running                  */

/* dev is the device                                                     */
/* prm is the p.e string                                                 */
/* state is TRUE or FALSE for setting on or off                          */

int send_indi_switch (FILE *fp, char *dev, char *prm, int state)
{
  char  p[1024], e[2048];
  
  /* Parse parameter string to components */
  
  sscanf (prm, "%[^.].%[^.]",  p, e);
  
  /* Create and send an XML string                        */
  
  /* Sample format:                                       */
  /*                                                      */
  /* setINDI ccd.ImageBinning.2=On                        */
  /*                                                      */
  /* <newSwitchVector device='ccd' name='ImageBinning'>   */
  /* <oneSwitch name='2'>'On'</oneSwitch>                 */
  /* </oneSwitchVector>                                   */
  
  fprintf (fp, "<newSwitchVector device='%s' name='%s'>\n",dev,p);
  if (state == TRUE)
  {
    fprintf (fp, "  <oneSwitch name='%s'>On</oneSwitch>\n",e);
  }
  else
  {
    fprintf (fp, "  <oneSwitch name='%s'>Off</oneSwitch>\n",e);
  }
  fprintf (fp, "</newSwitchVector>\n");
  fflush (fp);
  if (feof(fp) || ferror(fp)) 
  {
    fprintf (stderr, "Send error\n");
    return 1;
  }
  return 0;
}

/* int send_indi_text (char *dev, char *prm, char *txt)                     */

/* Derived from the command line programs get/setINDI.c by Elwood Downey */

/* Connects to an INDI server                                            */
/* Sends  property.elements=value to a device                            */
/* Returns (0) success,  (1) send error, (2)  trouble opening server     */
/* May fail without notice if the server is not running                  */

/* dev is the device                                                     */
/* prm is the p.e string                                                 */
/* txt is the new string value                                           */


int send_indi_text (FILE *fp, char *dev, char *prm, char *txt)
{
  char  p[1024], e[2048];
  
  /* Parse parameter string to components */
  
  sscanf (prm, "%[^.].%[^.]",  p, e);
  
  /* Create and send an XML string                        */
  
  /* Sample format:                                       */
  /*                                                      */
  /* setINDI ccd.File.BaseName=newname                    */
  /*                                                      */
  /* <newTextVector device='ccd' name='File'>             */
  /* <oneText name='BaseName'>'newname'</oneText>         */
  /* </oneTextVector>                                     */
  
  fprintf (fp, "<newTextVector device='%s' name='%s'>\n",dev,p);
  fprintf (fp, "  <oneText name='%s'>%s</oneText>\n",e,txt);
  fprintf (fp, "</newTextVector>\n");
  fflush (fp);
  if (feof(fp) || ferror(fp)) 
  {
    fprintf (stderr, "Send error\n");
    return 1;
  }
  return 0;
}

/* Responses to incoming INDI commands in TELCOMMANDLIST         */
/* These are already separated by type: number, switch, text ... */
/* XmTel only uses number and switch types                       */

/* Set a new value for an indi number property for the telescope */
/* This routine handles                                          */
/*   telescope.ra                                                */
/*   telescope.dec                                               */
/*   telescope_2000.ra                                           */
/*   telescope_2000.dec                                          */
/*   target.ra                                                   */
/*   target.dec                                                  */
/*   target_2000.ra                                              */
/*   target_2000.dec                                             */
/*   model_parameters.ha                                         */
/*   model_parameters.dec                                        */
/*   focus_reading.microns                                       */
/*   rotation_reading.degrees                                    */
/*   temperature_reading.celsius                                 */

static int set_indi_number (XMLEle *root, char errmsg[])
{
  XMLEle *ep;
  char *device;
  char *name;
  char *ep_name;
  
  device = findXMLAttValu (root, "device");
  if (strcmp(device,telescope_device) != 0)
  {
    return (0);
  }      
  
  name = findXMLAttValu (root, "name");
  if (strcmp(name,"telescope") == 0) 
  {
    for (ep = nextXMLEle (root, 1); ep != NULL; ep = nextXMLEle (root, 0))
    {
      ep_name = findXMLAttValu (ep, "name");
      if (strcmp(ep_name,"ra") == 0)
      {
        telra = atof(pcdataXMLEle(ep));
      }
      else if (strcmp(ep_name,"dec") == 0);
      {
        teldec = atof(pcdataXMLEle(ep));
      }
      show_telescope_coordinates();
    }
  }
  else if (strcmp(name,"telescope_2000") == 0) 
  {
    for (ep = nextXMLEle (root, 1); ep != NULL; ep = nextXMLEle (root, 0))
    {
      ep_name = findXMLAttValu (ep, "name");
      if (strcmp(ep_name,"ra") == 0)
      {
        telra2 = atof(pcdataXMLEle(ep));
      }
      else if (strcmp(ep_name,"dec") == 0);
      {
        teldec2 = atof(pcdataXMLEle(ep));
      }
      show_telescope_coordinates();
    }
  }
  else if (strcmp(name,"target") == 0) 
  {
    for (ep = nextXMLEle (root, 1); ep != NULL; ep = nextXMLEle (root, 0))
    {
      ep_name = findXMLAttValu (ep, "name");
      if (strcmp(ep_name,"ra") == 0)
      {
        targetra = atof(pcdataXMLEle(ep));
      }
      else if (strcmp(ep_name,"dec") == 0);
      {
        targetdec = atof(pcdataXMLEle(ep));
      }
      show_target_coordinates();
    }
  }
  else if (strcmp(name,"target_2000") == 0) 
  {
    for (ep = nextXMLEle (root, 1); ep != NULL; ep = nextXMLEle (root, 0))
    {
      ep_name = findXMLAttValu (ep, "name");
      if (strcmp(ep_name,"ra") == 0)
      {
        targetra2 = atof(pcdataXMLEle(ep));
      }
      else if (strcmp(ep_name,"dec") == 0);
      {
        targetdec2 = atof(pcdataXMLEle(ep));
      }
      show_target_coordinates();
    }
  }
  else if (strcmp(name,"model_parameters") == 0) 
  {
    for (ep = nextXMLEle (root, 1); ep != NULL; ep = nextXMLEle (root, 0))
    {
      ep_name = findXMLAttValu (ep, "name");
      if (strcmp(ep_name,"ha") == 0)
      {
        modelha1 = atof(pcdataXMLEle(ep));
      }
      else if (strcmp(ep_name,"dec") == 0);
      {
        modeldec1 = atof(pcdataXMLEle(ep));
      }
      show_model_parameters();
    }
  }  
  
  else if(strcmp(name,"focus_reading") == 0)
  {
    for (ep = nextXMLEle (root, 1); ep != NULL; ep = nextXMLEle (root, 0))
    {
      telfocus = atof(pcdataXMLEle(ep));
      show_focus();
    }
  }
  else if(strcmp(name,"rotation_reading") == 0)
  {
    for (ep = nextXMLEle (root, 1); ep != NULL; ep = nextXMLEle (root, 0))
    {
      telrotation = atof(pcdataXMLEle(ep));
      show_rotation();
    }
  }  
  else if(strcmp(name,"temperature_reading") == 0)
  {
    for (ep = nextXMLEle (root, 1); ep != NULL; ep = nextXMLEle (root, 0))
    {
      teltemperature = atof(pcdataXMLEle(ep));
      show_temperature();
    }
  }           
  else
  { 
    /* XmTel received something unexpected */

    if (diagnostics >= 1)
    {   
      fprintf(stderr,"XmTel: Unexpected INDI number %s\n",name);
      for (ep = nextXMLEle (root, 1); ep != NULL; ep = nextXMLEle (root, 0))
        fprintf (stderr,"  %s\n", pcdataXMLEle(ep));
    }
  }  
  
  if (diagnostics == 2)
  {
    fprintf(stderr,"Telescope EOD  ra dec = %f %f\n",
      telra, teldec);
    fprintf(stderr,"Telescope 2000.0  ra dec = %f %f\n",
      telra2, teldec2); 
    fprintf(stderr,"Target EOD  ra dec = %f %f\n",
      targetra, targetdec);
    fprintf(stderr,"Target 2000.0  ra dec = %f %f\n",
      targetra2, targetdec2);            
    fprintf(stderr,"Focus = %f\n", telfocus);
    fprintf(stderr,"Rotation = %f\n", telrotation);
    fprintf(stderr,"Temperature = %f\n", teltemperature);
  }
  return (0);
}

/* Set a new value for an indi switch property for the telescope */
/* This routine handles                                          */
/*    pointing.offset                                      */
/*    pointing.refraction                                  */
/*    pointing.polar                                       */
/*    pointing.dynamic                                     */
/*    reference.target                                           */
/*    reference.wcs                                              */
/*    reference.clear                                            */
/*    reference.save                                             */
/*    reference.recall                                           */
/*    reference.default                                          */
/*    goto.target                                                */
/*    slew.north                                                 */
/*    slew.south                                                 */
/*    slew.east                                                  */
/*    slew.west                                                  */
/*    slew_speed.slew                                            */
/*    slew_speed.find                                            */
/*    slew_speed.center                                          */
/*    slew_speed.guide                                           */
/*    track.on                                                   */
/*    guide.on                                                   */
/*    focus.off                                                  */
/*    focus.in                                                   */
/*    focus.out                                                  */
/*    focus_speed.slow                                           */
/*    focus_speed.fast                                           */
/*    rotate.off                                                 */
/*    rotate.ccw                                                 */
/*    rotate.cw                                                  */
/*    rotate_speed.sidereal                                      */
/*    rotate_speed.slow                                          */
/*    rotate_speed.fast                                          */
/*    fan.off                                                    */
/*    fan.low                                                    */
/*    fan.high                                                   */
/*    heater.off                                                 */
/*    heater.low                                                 */
/*    heater.high                                                */
                                                                                                                               
static int set_indi_switch (XMLEle *root, char errmsg[])        
{                                                               
  XMLEle *ep;                                                   
  char *device;
  char *name;                                                   
  char *ep_name;

  device = findXMLAttValu (root, "device");
  if (strcmp(device,telescope_device) != 0)
  {
    return (0);
  }      
                                                
  name = findXMLAttValu (root, "name");
  if (strcmp(name,"pointing_model") == 0) 
  {
    for (ep = nextXMLEle (root, 1); ep != NULL; ep = nextXMLEle (root, 0))
    {
      ep_name = findXMLAttValu (ep, "name");
      if (strcmp(ep_name,"offset") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {       
          pmodel=(pmodel | OFFSET);         
          XmToggleButtonSetState(p_options_toggle_1, True, False);
        }
        else if (strcmp(pcdataXMLEle(ep),"Off") == 0)
        {       
          pmodel=(pmodel & ~(OFFSET));         
          XmToggleButtonSetState(p_options_toggle_1, False, False);
        }
      }
      else if (strcmp(ep_name,"refraction") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {          
          pmodel=(pmodel | REFRACT);         
          XmToggleButtonSetState(p_options_toggle_2, True, False);
        }
        else if (strcmp(pcdataXMLEle(ep),"Off") == 0)
        {       
          pmodel=(pmodel & ~(REFRACT));         
          XmToggleButtonSetState(p_options_toggle_2, False, False);
        }      
      }
      else if (strcmp(ep_name,"polar") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {          
          pmodel=(pmodel | POLAR);         
          XmToggleButtonSetState(p_options_toggle_4, True, False);
        }
        else if (strcmp(pcdataXMLEle(ep),"Off") == 0)
        {       
          pmodel=(pmodel & ~(POLAR));         
          XmToggleButtonSetState(p_options_toggle_4, False, False);
        }      
      }
      else if (strcmp(ep_name,"dynamic") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {          
          pmodel=(pmodel | DYNAMIC);         
          XmToggleButtonSetState(p_options_toggle_8, True, False);
        }
        else if (strcmp(pcdataXMLEle(ep),"Off") == 0)
        {       
          pmodel=(pmodel & ~(DYNAMIC));         
          XmToggleButtonSetState(p_options_toggle_8, False, False);
        }        
      }
    }  
  }  
  else if (strcmp(name,"goto") == 0) 
  {
    for (ep = nextXMLEle (root, 1); ep != NULL; ep = nextXMLEle (root, 0))
    {
      if (strcmp(pcdataXMLEle(ep),"Off") == 0)
      {
        if (gotocmd == TRUE)
        {
          strcpy(message,"Go to target completed\n");      
          show_message();
        }
        gotocmd = FALSE;        
      }
      else if (strcmp(pcdataXMLEle(ep),"On") == 0)
      {
        if (gotocmd == FALSE)
        {
          strcpy(message,"Go to target started\n");      
          show_message();        
        }
        gotocmd = TRUE;
      }
    }
  }
  else if (strcmp(name,"set_target") == 0) 
  {
    targetcmd = TARGETCMDOFF; 
    for (ep = nextXMLEle (root, 1); ep != NULL; ep = nextXMLEle (root, 0))
    {
      ep_name = findXMLAttValu (ep, "name");
      if (strcmp(ep_name,"meridian") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          targetcmd = TARGETCMDMERIDIAN;
        }
      }
      else if (strcmp(ep_name,"zenith") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          targetcmd = TARGETCMDZENITH;
        }
      }
      else if (strcmp(ep_name,"park") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          targetcmd = TARGETCMDPARK;
        }
      }
    }
  }    
  else if (strcmp(name,"slew") == 0) 
  {
    slewcmd = FALSE;
    teldir = FALSE;
    for (ep = nextXMLEle (root, 1); ep != NULL; ep = nextXMLEle (root, 0))
    {
      ep_name = findXMLAttValu (ep, "name");
      if (strcmp(ep_name,"north") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          slewcmd = TRUE;
          teldir = NORTH;
        }
      }
      else if (strcmp(ep_name,"south") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          slewcmd = TRUE;
          teldir = SOUTH;
        }
      }
      else if (strcmp(ep_name,"east") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          slewcmd = TRUE;
          teldir = EAST;
        }
      }
      else if (strcmp(ep_name,"west") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          slewcmd = TRUE;
          teldir = WEST;
        }
      }           
    }
  }
  else if (strcmp(name,"slew_speed") == 0) 
  {
    for (ep = nextXMLEle (root, 1); ep != NULL; ep = nextXMLEle (root, 0))
    {
      ep_name = findXMLAttValu (ep, "name");
      if (strcmp(ep_name,"slew") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          telspd = SLEW;
          XmToggleButtonSetState(toggle[0], True, False);
          XmToggleButtonSetState(toggle[1], False, False);
          XmToggleButtonSetState(toggle[2], False, False);
          XmToggleButtonSetState(toggle[3], False, False);
        }
      }
      else if (strcmp(ep_name,"find") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          telspd = FIND;
          XmToggleButtonSetState(toggle[0], False, False);
          XmToggleButtonSetState(toggle[1], True, False);
          XmToggleButtonSetState(toggle[2], False, False);
          XmToggleButtonSetState(toggle[3], False, False);
        }
      }
      else if (strcmp(ep_name,"center") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          telspd = CENTER;
          XmToggleButtonSetState(toggle[0], False, False);
          XmToggleButtonSetState(toggle[1], False, False);
          XmToggleButtonSetState(toggle[2], True, False);
          XmToggleButtonSetState(toggle[3], False, False);
        }
      }
      else if (strcmp(ep_name,"guide") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          telspd = GUIDE;
          XmToggleButtonSetState(toggle[0], False, False);
          XmToggleButtonSetState(toggle[1], False, False);
          XmToggleButtonSetState(toggle[2], False, False);
          XmToggleButtonSetState(toggle[3], True, False);
        }
      }           
    }
  }
  else if (strcmp(name,"track") == 0) 
  {
    for (ep = nextXMLEle (root, 1); ep != NULL; ep = nextXMLEle (root, 0))
    {
      if (strcmp(pcdataXMLEle(ep),"Off") == 0)
      {
        if ( trackcmd == TRUE )
        {
          strcpy(message,"Tracking off\n");      
          show_message();
        }
        trackcmd = FALSE;        
      }
      else if (strcmp(pcdataXMLEle(ep),"On") == 0)
      {
        if ( trackcmd == FALSE )
        {
          strcpy(message,"Tracking on\n");      
          show_message();
        }        
        trackcmd = TRUE;
      }
    }
  }
  else if (strcmp(name,"guide") == 0) 
  {
    for (ep = nextXMLEle (root, 1); ep != NULL; ep = nextXMLEle (root, 0))
    {
      if (strcmp(pcdataXMLEle(ep),"Off") == 0)
      {
        if ( guidecmd == TRUE )
        {
          strcpy(message,"Guiding off\n");      
          show_message();
        }
        guidecmd = FALSE;        
      }
      else if (strcmp(pcdataXMLEle(ep),"On") == 0)
      {
        if ( guidecmd == FALSE )
        {
          strcpy(message,"Guiding on\n");      
          show_message();
        }        
        guidecmd = TRUE;
      }
    }
  }  
  else if (strcmp(name,"focus") == 0) 
  {
    for (ep = nextXMLEle (root, 1); ep != NULL; ep = nextXMLEle (root, 0))
    {
      ep_name = findXMLAttValu (ep, "name");
      if (strcmp(ep_name,"off") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          focuscmd = FOCUSCMDOFF;
        }
      }
      else if (strcmp(ep_name,"in") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
         focuscmd = FOCUSCMDIN;
        }
      }
      else if (strcmp(ep_name,"out") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
         focuscmd = FOCUSCMDOUT;
        }
      }
    }
  }
  else if (strcmp(name,"focus_speed") == 0) 
  {
    for (ep = nextXMLEle (root, 1); ep != NULL; ep = nextXMLEle (root, 0))
    {
      ep_name = findXMLAttValu (ep, "name");
      if (strcmp(ep_name,"8x") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          focusspd = FOCUSSPD4;
          XmToggleButtonSetState(focus_spd_toggle[0], True, False);
          XmToggleButtonSetState(focus_spd_toggle[1], False, False);
          XmToggleButtonSetState(focus_spd_toggle[2], False, False);
          XmToggleButtonSetState(focus_spd_toggle[3], False, False);
        }
      }
      else if (strcmp(ep_name,"4x") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          focusspd = FOCUSSPD3;
          XmToggleButtonSetState(focus_spd_toggle[0], False, False);
          XmToggleButtonSetState(focus_spd_toggle[1], True, False);
          XmToggleButtonSetState(focus_spd_toggle[2], False, False);
          XmToggleButtonSetState(focus_spd_toggle[3], False, False);
        }
      }
      else if (strcmp(ep_name,"2x") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          focusspd = FOCUSSPD2;
          XmToggleButtonSetState(focus_spd_toggle[0], False, False);
          XmToggleButtonSetState(focus_spd_toggle[1], False, False);
          XmToggleButtonSetState(focus_spd_toggle[2], True, False);
          XmToggleButtonSetState(focus_spd_toggle[3], False, False);
        }
      }
      else if (strcmp(ep_name,"1x") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          focusspd = FOCUSSPD1;
          XmToggleButtonSetState(focus_spd_toggle[0], False, False);
          XmToggleButtonSetState(focus_spd_toggle[1], False, False);
          XmToggleButtonSetState(focus_spd_toggle[2], False, False);
          XmToggleButtonSetState(focus_spd_toggle[3], True, False);
        }
      }    
    }
  }
  else if (strcmp(name,"rotate") == 0) 
  {
    for (ep = nextXMLEle (root, 1); ep != NULL; ep = nextXMLEle (root, 0))
    {
      ep_name = findXMLAttValu (ep, "name");
      if (strcmp(ep_name,"off") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          rotatecmd = ROTATECMDOFF;
          XmToggleButtonSetState(rotate_cmd_toggle[0], False, False);
          XmToggleButtonSetState(rotate_cmd_toggle[1], False, False);
          XmToggleButtonSetState(rotate_cmd_toggle[2], True, False);          
        }
      }
      else if (strcmp(ep_name,"cw") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          rotatecmd = ROTATECMDCW;
          XmToggleButtonSetState(rotate_cmd_toggle[0], True, False);
          XmToggleButtonSetState(rotate_cmd_toggle[1], False, False);
          XmToggleButtonSetState(rotate_cmd_toggle[2], False, False);          
        }
      }
      else if (strcmp(ep_name,"ccw") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          rotatecmd = ROTATECMDCCW;
          XmToggleButtonSetState(rotate_cmd_toggle[0], False, False);
          XmToggleButtonSetState(rotate_cmd_toggle[1], True, False);
          XmToggleButtonSetState(rotate_cmd_toggle[2], False, False);          
        }
      }
    }
  }
  else if (strcmp(name,"rotate_speed") == 0) 
  {
    for (ep = nextXMLEle (root, 1); ep != NULL; ep = nextXMLEle (root, 0))
    {
      ep_name = findXMLAttValu (ep, "name");
      if (strcmp(ep_name,"fast") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          rotatespd = ROTATESPDSIDEREAL;
          XmToggleButtonSetState(rotate_spd_toggle[0], True, False);
          XmToggleButtonSetState(rotate_spd_toggle[1], False, False);
          XmToggleButtonSetState(rotate_spd_toggle[2], False, False);
         }
      }
      else if (strcmp(ep_name,"slow") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          rotatespd = ROTATESPDSLOW;
          XmToggleButtonSetState(rotate_spd_toggle[0], False, False);
          XmToggleButtonSetState(rotate_spd_toggle[1], True, False);
          XmToggleButtonSetState(rotate_spd_toggle[2], False, False);
        }
      }
      else if (strcmp(ep_name,"sidereal") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          rotatespd = ROTATESPDFAST;
          XmToggleButtonSetState(rotate_spd_toggle[0], False, False);
          XmToggleButtonSetState(rotate_spd_toggle[1], False, False);
          XmToggleButtonSetState(rotate_spd_toggle[2], True, False);
        }
      }      
    }
  }
  else if (strcmp(name,"fan") == 0) 
  {
    for (ep = nextXMLEle (root, 1); ep != NULL; ep = nextXMLEle (root, 0))
    {
      ep_name = findXMLAttValu (ep, "name");
      if (strcmp(ep_name,"high") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          fancmd = FANCMDOFF;
          XmToggleButtonSetState(fan_spd_toggle[0], True, False);
          XmToggleButtonSetState(fan_spd_toggle[1], False, False);
          XmToggleButtonSetState(fan_spd_toggle[2], False, False);
        }
      }
      else if (strcmp(ep_name,"low") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          fancmd = FANCMDLOW;
          XmToggleButtonSetState(fan_spd_toggle[0], False, False);
          XmToggleButtonSetState(fan_spd_toggle[1], True, False);
          XmToggleButtonSetState(fan_spd_toggle[2], False, False);
        }
      }
      else if (strcmp(ep_name,"off") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          fancmd = FANCMDHIGH;
          XmToggleButtonSetState(fan_spd_toggle[0], False, False);
          XmToggleButtonSetState(fan_spd_toggle[1], False, False);
          XmToggleButtonSetState(fan_spd_toggle[2], True, False);
        }
      }      
    }
  }
  else if (strcmp(name,"heater") == 0) 
  {
    for (ep = nextXMLEle (root, 1); ep != NULL; ep = nextXMLEle (root, 0))
    {
      ep_name = findXMLAttValu (ep, "name");
      if (strcmp(ep_name,"high") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          heatercmd = HEATERCMDOFF;
          XmToggleButtonSetState(heater_pwr_toggle[0], True, False);
          XmToggleButtonSetState(heater_pwr_toggle[1], False, False);
          XmToggleButtonSetState(heater_pwr_toggle[2], False, False);
        }
      }
      else if (strcmp(ep_name,"low") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          heatercmd = HEATERCMDLOW;
          XmToggleButtonSetState(heater_pwr_toggle[0], False, False);
          XmToggleButtonSetState(heater_pwr_toggle[1], True, False);
          XmToggleButtonSetState(heater_pwr_toggle[2], False, False);
        }
      }
      else if (strcmp(ep_name,"off") == 0)
      {
        if (strcmp(pcdataXMLEle(ep),"On") == 0)
        {
          heatercmd = HEATERCMDHIGH;
          XmToggleButtonSetState(heater_pwr_toggle[0], False, False);
          XmToggleButtonSetState(heater_pwr_toggle[1], False, False);
          XmToggleButtonSetState(heater_pwr_toggle[2], True, False);
        }
      }      
    }
  }
  else
  {
    
    /* XmTel received something unexpected */

    if (diagnostics >= 1)
    {
      fprintf(stderr,"XmTel: Unexpected INDI switch %s\n",name);
      for (ep = nextXMLEle (root, 1); ep != NULL; ep = nextXMLEle (root, 0))
        fprintf (stderr,"  %s\n", pcdataXMLEle(ep));
    }
  }

  /* More diagnostic messages */

  if (diagnostics == 2)
  {
    fprintf(stderr,"telflag = %d\n",  telflag);
    fprintf(stderr,"pmodel = %d\n",   pmodel);
    fprintf(stderr,"refcmd = %d\n",   refcmd); 
    fprintf(stderr,"telspd = %d\n",   telspd);
    fprintf(stderr,"teldir = %d\n",   teldir);
    fprintf(stderr,"trackcmd = %d\n", trackcmd);
    fprintf(stderr,"gotocmd = %d\n",  gotocmd);
    fprintf(stderr,"guidecmd = %d\n", guidecmd);
    fprintf(stderr,"slewcmd = %d\n",  slewcmd);
    fprintf(stderr,"focuscmd = %d\n", focuscmd);
    fprintf(stderr,"focusspd = %d\n", focusspd);
  } 
  return (0);
}

/* Set a new value for an indi text property for the telescope*/

static int set_indi_text (XMLEle *root, char errmsg[])
{
  char *device;
  char *name;
  device = findXMLAttValu (root, "device");
  if (strcmp(device,telescope_device) == 0)
  {
    name = findXMLAttValu (root, "name");
    fprintf(stderr,"Unexpected inditel text %s \n",name);
  }
  return (0);
}

/* Set a new value for an indi light property for the telescope */

static int set_indi_light (XMLEle *root, char errmsg[])
{
  char *device;
  char *name;
  device = findXMLAttValu (root, "device");
  if (strcmp(device,telescope_device) == 0)
  {  
    name = findXMLAttValu (root, "name");
    fprintf(stderr,"Unexpected inditel light %s \n",name);
  } 
  return (0);  
}

/* Set a new value for an indi blob property for the telescope */
static int set_indi_blob (XMLEle *root, char errmsg[])
{  
  char *device;
  char *name;
  device = findXMLAttValu (root, "device");
  if (strcmp(device,telescope_device) == 0)
  {    
    name = findXMLAttValu (root, "name");
    fprintf(stderr,"Unexpected inditel blob %s \n",name);
  }
  return (0);
}

/* Display a message from the server */
static int set_indi_message (XMLEle *root, char errmsg[])
{
  fprintf(stderr,"Indi message: %s\n",errmsg);
  return (0);
}
